# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
An interactive phase-plane plot generated from a Model object of TVB.

Optionally an Integrator object from TVB can be specified, this will be used to
generate sample trajectories -- not the phase-plane. This is mainly interesting
for visualising the effect of noise on a trajectory.

Demo::

    from tvb.simulator.plot.phase_plane_interactive import PhasePlaneInteractive
    ppi_fig = PhasePlaneInteractive()
    ppi_fig.show()


Example specifying a Model and stochastic sample trajectories::

    import tvb.simulator.models
    from tvb.simulator.plot.phase_plane_interactive import PhasePlaneInteractive
    MODEL = tvb.simulator.models.JansenRit()
    import tvb.simulator.integrators
    INTEGRATOR = tvb.simulator.integrators.HeunStochastic(dt=2**-5)
    ppi_fig = PhasePlaneInteractive(model=MODEL, integrator=INTEGRATOR)
    ppi_fig.show()


.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""
# TODO: Add more LOG statements.
# TODO: Add connectivity term to parameters...
# TODO: Memory grows with usage, may be just a lazy garbage collector but, should
#      check for leaks or look into "forcing" cleanup...

import numpy
import matplotlib.pyplot as plt
import colorsys
import numbers
import matplotlib.widgets as widgets
from deprecated import deprecated
from tvb.simulator.common import get_logger
import tvb.simulator.models as models_module
import tvb.simulator.integrators as integrators_module
from tvb.basic.neotraits.api import HasTraits, Attr, NArray, List

LOG = get_logger(__name__)

# Define a colour theme... see: matplotlib.colors.cnames.keys()
BACKGROUNDCOLOUR = "lightgray"
EDGECOLOUR = "darkslateblue"
AXCOLOUR = "steelblue"
BUTTONCOLOUR = "steelblue"
HOVERCOLOUR = "darkred"

# Set the resolution of the phase-plane and sample trajectories.
NUMBEROFGRIDPOINTS = 42
TRAJ_STEPS = 4096


def get_color(num_colours):
    for hue in range(num_colours):
        hue = 1.0 * hue / num_colours
        col = [int(x) for x in colorsys.hsv_to_rgb(hue, 1.0, 230)]
        yield "#{0:02x}{1:02x}{2:02x}".format(*col)


@deprecated(reason="Use tvb-widgets instead")
class PhasePlaneInteractive(HasTraits):
    """
    The GUI for the interactive phase-plane viewer provides sliders for setting:
        - The value of all parameters of the Model.
        - The extent of the axes.
        - A fixed value for the state-variables which aren't currently selected.
        - The noise strength, if a stocahstic integrator is specified.

    and radio buttons for selecting:
        - Which state-variables to show on each axis.
        - Which mode to show, if the Model has them.

    Clicking on the phase-plane will generate a sample trajectory, originating
    from where you clicked.

    """

    model = Attr(
        field_type=models_module.Model,
        label="Model",
        default=models_module.Generic2dOscillator(),
        doc="""An instance of the local dynamic model to be investigated with
        PhasePlaneInteractive.""")

    integrator = Attr(
        field_type=integrators_module.Integrator,
        label="Integrator",
        default=integrators_module.RungeKutta4thOrderDeterministic(),
        doc="""The integration scheme used to for generating sample
        trajectories on the phase-plane. NOTE: This is not used for generating
        the phase-plane itself, ie the vector field and nulclines.""")

    exclude_sliders = List(
        of=str)

    def __init__(self, **kwargs):
        """
        Initialise based on provided keywords or their traited defaults. Also,
        initialise the place-holder attributes that aren't filled until the
        show() method is called.
        """
        super(PhasePlaneInteractive, self).__init__(**kwargs)
        LOG.debug(str(kwargs))

        # figure
        self.ipp_fig = None

        # phase-plane
        self.pp_ax = None
        self.X = None
        self.Y = None
        self.U = None
        self.V = None
        self.UVmag = None
        self.nullcline_x = None
        self.nullcline_y = None
        self.pp_quivers = None

        # Current state
        self.svx = None
        self.svy = None
        self.default_sv = None
        self.no_coupling = None
        self.mode = None

        # Selectors
        self.state_variable_x = None
        self.state_variable_y = None
        self.mode_selector = None

        # Sliders
        self.param_sliders = None
        self.axes_range_sliders = None
        self.sv_sliders = None
        self.noise_slider = None

        # Reset buttons
        self.reset_param_button = None
        self.reset_sv_button = None
        self.reset_axes_button = None
        self.reset_noise_button = None
        self.reset_seed_button = None

    def show(self):
        """ Generate the interactive phase-plane figure. """
        model_name = self.model.__class__.__name__
        msg = "Generating an interactive phase-plane plot for %s"
        LOG.info(msg % model_name)

        # Make sure the model is fully configured...
        self.model.configure()

        # Setup the inital(current) state
        self.svx = self.model.state_variables[0]  # x-axis: 1st state variable
        self.svy = self.model.state_variables[1]  # y-axis: 2nd state variable
        self.mode = 0
        self.set_state_vector()

        # Make the figure:
        self.create_figure()

        # Selectors
        self.add_state_variable_selector()
        self.add_mode_selector()

        # Sliders
        self.add_axes_range_sliders()
        self.add_state_variable_sliders()
        self.add_param_sliders()
        if isinstance(self.integrator, integrators_module.IntegratorStochastic):
            if self.integrator.noise.ntau > 0.0:
                self.integrator.noise.configure_coloured(self.integrator.dt,
                                                         (1, self.model.nvar, 1,
                                                          self.model.number_of_modes))
            else:
                self.integrator.noise.configure_white(self.integrator.dt,
                                                      (1, self.model.nvar, 1,
                                                       self.model.number_of_modes))

            self.add_noise_slider()
            self.add_reset_noise_button()
            self.add_reset_seed_button()

        # Reset buttons
        self.add_reset_param_button()
        self.add_reset_sv_button()
        self.add_reset_axes_button()

        # Calculate the phase plane
        self.set_mesh_grid()
        self.calc_phase_plane()

        # Plot phase plane
        self.plot_phase_plane()

        # add mouse handler for trajectory clicking
        self.ipp_fig.canvas.mpl_connect('button_press_event',
                                        self.click_trajectory)
        # import pdb; pdb.set_trace()

        plt.show()

    ##------------------------------------------------------------------------##
    ##----------------- Functions for building the figure --------------------##
    ##------------------------------------------------------------------------##

    def create_figure(self):
        """ Create the figure and phase-plane axes. """
        # Figure and main phase-plane axes
        model_name = self.model.__class__.__name__
        integrator_name = self.integrator.__class__.__name__
        figsize = 10, 5
        try:
            figure_window_title = "Interactive phase-plane: " + model_name
            figure_window_title += "   --   %s" % integrator_name
            self.ipp_fig = plt.figure(num=figure_window_title,
                                      figsize=figsize,
                                      facecolor=BACKGROUNDCOLOUR,
                                      edgecolor=EDGECOLOUR)
        except ValueError:
            LOG.info("My life would be easier if you'd update your PyLab...")
            self.ipp_fig = plt.figure(num=42, figsize=figsize,
                                      facecolor=BACKGROUNDCOLOUR,
                                      edgecolor=EDGECOLOUR)

        self.pp_ax = self.ipp_fig.add_axes([0.265, 0.2, 0.5, 0.75])

        self.pp_splt = self.ipp_fig.add_subplot(212)
        self.ipp_fig.subplots_adjust(left=0.265, bottom=0.02, right=0.765,
                                     top=0.3, wspace=0.1, hspace=None)
        self.pp_splt.set_prop_cycle(color=get_color(self.model.nvar))
        self.pp_splt.plot(numpy.arange(TRAJ_STEPS + 1) * self.integrator.dt,
                          numpy.zeros((TRAJ_STEPS + 1, self.model.nvar)))
        if hasattr(self.pp_splt, 'autoscale'):
            self.pp_splt.autoscale(enable=True, axis='y', tight=True)
        self.pp_splt.legend(self.model.state_variables)

    def add_state_variable_selector(self):
        """
        Generate radio selector buttons to set which state variable is displayed
        on the x and y axis of the phase-plane plot.
        """
        svx_ind = self.model.state_variables.index(self.svx)
        svy_ind = self.model.state_variables.index(self.svy)

        # State variable for the x axis
        pos_shp = [0.07, 0.05, 0.065, 0.12 + 0.006 * self.model.nvar]
        rax = self.ipp_fig.add_axes(pos_shp, facecolor=AXCOLOUR, title="x-axis")
        self.state_variable_x = widgets.RadioButtons(rax,
                                                     tuple(self.model.state_variables),
                                                     active=svx_ind)
        self.state_variable_x.on_clicked(self.update_svx)

        # State variable for the y axis
        pos_shp = [0.14, 0.05, 0.065, 0.12 + 0.006 * self.model.nvar]
        rax = self.ipp_fig.add_axes(pos_shp, facecolor=AXCOLOUR, title="y-axis")
        self.state_variable_y = widgets.RadioButtons(rax,
                                                     tuple(self.model.state_variables),
                                                     active=svy_ind)
        self.state_variable_y.on_clicked(self.update_svy)

    def add_mode_selector(self):
        """
        Add a radio button to the figure for selecting which mode of the model
        should be displayed.
        """
        pos_shp = [0.02, 0.07, 0.04, 0.1 + 0.002 * self.model.number_of_modes]
        rax = self.ipp_fig.add_axes(pos_shp, facecolor=AXCOLOUR, title="Mode")
        mode_tuple = tuple(range(self.model.number_of_modes))
        self.mode_selector = widgets.RadioButtons(rax, mode_tuple, active=0)
        self.mode_selector.on_clicked(self.update_mode)

    def add_axes_range_sliders(self):
        """
        Add sliders to the figure to allow the phase-planes axes to be set.
        """
        self.axes_range_sliders = dict()

        default_range_x = (self.model.state_variable_range[self.svx][1] -
                           self.model.state_variable_range[self.svx][0])
        default_range_y = (self.model.state_variable_range[self.svy][1] -
                           self.model.state_variable_range[self.svy][0])
        min_val_x = self.model.state_variable_range[self.svx][0] - 4.0 * default_range_x
        max_val_x = self.model.state_variable_range[self.svx][1] + 4.0 * default_range_x
        min_val_y = self.model.state_variable_range[self.svy][0] - 4.0 * default_range_y
        max_val_y = self.model.state_variable_range[self.svy][1] + 4.0 * default_range_y

        sax = self.ipp_fig.add_axes([0.04, 0.835, 0.125, 0.025],
                                    facecolor=AXCOLOUR)
        sl_x_min = widgets.Slider(sax, "xlo", min_val_x, max_val_x,
                                  valinit=self.model.state_variable_range[self.svx][0])
        sl_x_min.on_changed(self.update_range)

        sax = self.ipp_fig.add_axes([0.04, 0.8, 0.125, 0.025], facecolor=AXCOLOUR)
        sl_x_max = widgets.Slider(sax, "xhi", min_val_x, max_val_x,
                                  valinit=self.model.state_variable_range[self.svx][1])
        sl_x_max.on_changed(self.update_range)

        sax = self.ipp_fig.add_axes([0.04, 0.765, 0.125, 0.025],
                                    facecolor=AXCOLOUR)
        sl_y_min = widgets.Slider(sax, "ylo", min_val_y, max_val_y,
                                  valinit=self.model.state_variable_range[self.svy][0])
        sl_y_min.on_changed(self.update_range)

        sax = self.ipp_fig.add_axes([0.04, 0.73, 0.125, 0.025], facecolor=AXCOLOUR)
        sl_y_max = widgets.Slider(sax, "yhi", min_val_y, max_val_y,
                                  valinit=self.model.state_variable_range[self.svy][1])
        sl_y_max.on_changed(self.update_range)

        self.axes_range_sliders["sl_x_min"] = sl_x_min
        self.axes_range_sliders["sl_x_max"] = sl_x_max
        self.axes_range_sliders["sl_y_min"] = sl_y_min
        self.axes_range_sliders["sl_y_max"] = sl_y_max

    def add_state_variable_sliders(self):
        """
        Add sliders to the figure to allow default values for the models state
        variable to be set.
        """
        msv_range = self.model.state_variable_range
        offset = 0.0
        self.sv_sliders = dict()
        for sv in range(self.model.nvar):
            offset += 0.035
            pos_shp = [0.04, 0.6 - offset, 0.125, 0.025]
            sax = self.ipp_fig.add_axes(pos_shp, facecolor=AXCOLOUR)
            sv_str = self.model.state_variables[sv]
            self.sv_sliders[sv_str] = widgets.Slider(sax, sv_str,
                                                     msv_range[sv_str][0],
                                                     msv_range[sv_str][1],
                                                     valinit=self.default_sv[sv, 0, 0])
            self.sv_sliders[sv_str].on_changed(self.update_state_variables)

    # Traited paramaters as sliders
    def add_param_sliders(self):
        """
        Add sliders to the figure to allow the models parameters to be set.
        """
        offset = 0.0
        self.param_sliders = dict()
        # import pdb; pdb.set_trace()
        for param_name in type(self.model).declarative_attrs:
            if self.exclude_sliders is not None and param_name in self.exclude_sliders:
                continue
            param_def = getattr(type(self.model), param_name)
            if not isinstance(param_def, NArray) or not issubclass(param_def.dtype.type, numbers.Real):
                continue
            param_range = param_def.domain
            if param_range is None:
                continue
            offset += 0.035
            sax = self.ipp_fig.add_axes([0.825, 0.865 - offset, 0.125, 0.025],
                                        facecolor=AXCOLOUR)
            param_value = getattr(self.model, param_name)
            param_value = param_value[0] if len(param_value) > 0 else 0
            self.param_sliders[param_name] = widgets.Slider(sax, param_name,
                                                            param_range.lo,
                                                            param_range.hi,
                                                            valinit=param_value)
            self.param_sliders[param_name].on_changed(self.update_parameters)

    def add_noise_slider(self):
        """
        Add a slider to the figure to allow the integrators noise strength to
        be set.
        """
        pos_shp = [0.825, 0.1, 0.125, 0.025]
        sax = self.ipp_fig.add_axes(pos_shp, facecolor=AXCOLOUR)

        self.noise_slider = widgets.Slider(sax, "Log Noise", -9.0, 1.0,
                                           valinit=self.integrator.noise.nsig[0])
        self.noise_slider.on_changed(self.update_noise)

    def add_reset_param_button(self):
        """
        Add a button to the figure for reseting the model parameter values to
        their original values.
        """
        bax = self.ipp_fig.add_axes([0.825, 0.865, 0.125, 0.04])
        self.reset_param_button = widgets.Button(bax, 'Reset parameters',
                                                 color=BUTTONCOLOUR,
                                                 hovercolor=HOVERCOLOUR)

        def reset_parameters(event):
            for param_slider in self.param_sliders:
                self.param_sliders[param_slider].reset()

        self.reset_param_button.on_clicked(reset_parameters)

    def add_reset_sv_button(self):
        """
        Add a button to the figure for reseting the model state variables to
        their default values.
        """
        bax = self.ipp_fig.add_axes([0.04, 0.60, 0.125, 0.04])
        self.reset_sv_button = widgets.Button(bax, 'Reset state-variables',
                                              color=BUTTONCOLOUR,
                                              hovercolor=HOVERCOLOUR)

        def reset_state_variables(event):
            for svsl in self.sv_sliders.values():
                svsl.reset()

        self.reset_sv_button.on_clicked(reset_state_variables)

    def add_reset_noise_button(self):
        """
        Add a button to the figure for reseting the noise to its default value.
        """
        bax = self.ipp_fig.add_axes([0.825, 0.135, 0.125, 0.04])
        self.reset_noise_button = widgets.Button(bax, 'Reset noise strength',
                                                 color=BUTTONCOLOUR,
                                                 hovercolor=HOVERCOLOUR)

        def reset_noise(event):
            self.noise_slider.reset()

        self.reset_noise_button.on_clicked(reset_noise)

    def add_reset_seed_button(self):
        """
        Add a button to the figure for reseting the random number generator to
        its intial state. For reproducible noise...
        """
        bax = self.ipp_fig.add_axes([0.825, 0.05, 0.125, 0.04])
        self.reset_seed_button = widgets.Button(bax, 'Reset random stream',
                                                color=BUTTONCOLOUR,
                                                hovercolor=HOVERCOLOUR)

        def reset_seed(event):
            self.integrator.noise.trait["random_stream"].reset()

        self.reset_seed_button.on_clicked(reset_seed)

    def add_reset_axes_button(self):
        """
        Add a button to the figure for reseting the phase-plane axes to their
        default ranges.
        """
        bax = self.ipp_fig.add_axes([0.04, 0.87, 0.125, 0.04])
        self.reset_axes_button = widgets.Button(bax, 'Reset axes',
                                                color=BUTTONCOLOUR,
                                                hovercolor=HOVERCOLOUR)

        def reset_ranges(event):
            self.axes_range_sliders["sl_x_min"].reset()
            self.axes_range_sliders["sl_x_max"].reset()
            self.axes_range_sliders["sl_y_min"].reset()
            self.axes_range_sliders["sl_y_max"].reset()

        self.reset_axes_button.on_clicked(reset_ranges)

    ##------------------------------------------------------------------------##
    ##------------------- Functions for updating the figure ------------------##
    ##------------------------------------------------------------------------##

    # NOTE: All the ax.set_xlim, poly.xy, etc, garbage below is fragile. It works
    #      at the moment, but there are currently bugs in Slider and the hackery
    #      below takes these into account... If the bugs are fixed/changed then
    #      this could break. As an example, the Slider doc says poly is a
    #      Rectangle, but it's actually a Polygon. The Slider set_val method
    #      assumes a Rectangle even though this is not the case, so the array
    #      Slider.poly.xy is corrupted by that method. The corruption isn't
    #      visible in the plot, which is probably why it hasn't been fixed...

    def update_xrange_sliders(self):
        """
        A hacky update of the x-axis range sliders that is called when the
        state-variable selected for the x-axis is changed.
        """
        default_range_x = (self.model.state_variable_range[self.svx][1] -
                           self.model.state_variable_range[self.svx][0])
        min_val_x = self.model.state_variable_range[self.svx][0] - 4.0 * default_range_x
        max_val_x = self.model.state_variable_range[self.svx][1] + 4.0 * default_range_x
        self.axes_range_sliders["sl_x_min"].valinit = self.model.state_variable_range[self.svx][0]
        self.axes_range_sliders["sl_x_min"].valmin = min_val_x
        self.axes_range_sliders["sl_x_min"].valmax = max_val_x
        self.axes_range_sliders["sl_x_min"].ax.set_xlim(min_val_x, max_val_x)
        self.axes_range_sliders["sl_x_min"].poly.axes.set_xlim(min_val_x, max_val_x)
        self.axes_range_sliders["sl_x_min"].vline.set_data(
            ([self.axes_range_sliders["sl_x_min"].valinit, self.axes_range_sliders["sl_x_min"].valinit], [0, 1]))
        self.axes_range_sliders["sl_x_max"].valinit = self.model.state_variable_range[self.svx][1]
        self.axes_range_sliders["sl_x_max"].valmin = min_val_x
        self.axes_range_sliders["sl_x_max"].valmax = max_val_x
        self.axes_range_sliders["sl_x_max"].ax.set_xlim(min_val_x, max_val_x)
        self.axes_range_sliders["sl_x_max"].poly.axes.set_xlim(min_val_x, max_val_x)
        self.axes_range_sliders["sl_x_max"].vline.set_data(
            ([self.axes_range_sliders["sl_x_max"].valinit, self.axes_range_sliders["sl_x_max"].valinit], [0, 1]))
        self.axes_range_sliders["sl_x_min"].reset()
        self.axes_range_sliders["sl_x_max"].reset()

    def update_yrange_sliders(self):
        """
        A hacky update of the y-axis range sliders that is called when the
        state-variable selected for the y-axis is changed.
        """
        # svy_ind = self.model.state_variables.index(self.svy)
        default_range_y = (self.model.state_variable_range[self.svy][1] -
                           self.model.state_variable_range[self.svy][0])
        min_val_y = self.model.state_variable_range[self.svy][0] - 4.0 * default_range_y
        max_val_y = self.model.state_variable_range[self.svy][1] + 4.0 * default_range_y
        self.axes_range_sliders["sl_y_min"].valinit = self.model.state_variable_range[self.svy][0]
        self.axes_range_sliders["sl_y_min"].valmin = min_val_y
        self.axes_range_sliders["sl_y_min"].valmax = max_val_y
        self.axes_range_sliders["sl_y_min"].ax.set_xlim(min_val_y, max_val_y)
        self.axes_range_sliders["sl_y_min"].poly.axes.set_xlim(min_val_y, max_val_y)
        self.axes_range_sliders["sl_y_min"].vline.set_data(
            ([self.axes_range_sliders["sl_y_min"].valinit, self.axes_range_sliders["sl_y_min"].valinit], [0, 1]))
        self.axes_range_sliders["sl_y_max"].valinit = self.model.state_variable_range[self.svy][1]
        self.axes_range_sliders["sl_y_max"].valmin = min_val_y
        self.axes_range_sliders["sl_y_max"].valmax = max_val_y
        self.axes_range_sliders["sl_y_max"].ax.set_xlim(min_val_y, max_val_y)
        self.axes_range_sliders["sl_y_max"].poly.axes.set_xlim(min_val_y, max_val_y)
        self.axes_range_sliders["sl_y_max"].vline.set_data(
            ([self.axes_range_sliders["sl_y_max"].valinit, self.axes_range_sliders["sl_y_max"].valinit], [0, 1]))
        self.axes_range_sliders["sl_y_min"].reset()
        self.axes_range_sliders["sl_y_max"].reset()

    def update_svx(self, label):
        """
        Update state variable used for x-axis based on radio buttton selection.
        """
        self.svx = label
        self.update_xrange_sliders()
        self.set_mesh_grid()
        self.calc_phase_plane()
        self.update_phase_plane()

    def update_svy(self, label):
        """
        Update state variable used for y-axis based on radio buttton selection.
        """
        self.svy = label
        self.update_yrange_sliders()
        self.set_mesh_grid()
        self.calc_phase_plane()
        self.update_phase_plane()

    def update_mode(self, label):
        """ Update the visualised mode based on radio button selection. """
        self.mode = label
        self.update_phase_plane()

    def update_parameters(self, val):
        """
        Update model parameters based on the current parameter slider values.

        NOTE: Haven't figured out how to update independantly, so just update
            everything.
        """
        # TODO: Grab caller and use val directly, ie independent parameter update.
        # import pdb; pdb.set_trace()
        for param in self.param_sliders:
            setattr(self.model, param, numpy.array([self.param_sliders[param].val]))

        self.model.update_derived_parameters()
        self.calc_phase_plane()
        self.update_phase_plane()

    def update_noise(self, nsig):
        """ Update integrator noise based on the noise slider value. """
        self.integrator.noise.nsig = numpy.array([10 ** nsig, ])

    def update_range(self, val):
        """
        Update the axes ranges based on the current axes slider values.

        NOTE: Haven't figured out how to update independantly, so just update
            everything.

        """
        # TODO: Grab caller and use val directly, ie independent range update.
        self.axes_range_sliders["sl_x_min"].ax.set_facecolor(AXCOLOUR)
        self.axes_range_sliders["sl_x_max"].ax.set_facecolor(AXCOLOUR)
        self.axes_range_sliders["sl_y_min"].ax.set_facecolor(AXCOLOUR)
        self.axes_range_sliders["sl_y_max"].ax.set_facecolor(AXCOLOUR)

        if (self.axes_range_sliders["sl_x_min"].val >=
                self.axes_range_sliders["sl_x_max"].val):
            LOG.error("X-axis min must be less than max...")
            self.axes_range_sliders["sl_x_min"].ax.set_facecolor("Red")
            self.axes_range_sliders["sl_x_max"].ax.set_facecolor("Red")
            return
        if (self.axes_range_sliders["sl_y_min"].val >=
                self.axes_range_sliders["sl_y_max"].val):
            LOG.error("Y-axis min must be less than max...")
            self.axes_range_sliders["sl_y_min"].ax.set_facecolor("Red")
            self.axes_range_sliders["sl_y_max"].ax.set_facecolor("Red")
            return

        msv_range = self.model.state_variable_range
        msv_range[self.svx][0] = self.axes_range_sliders["sl_x_min"].val
        msv_range[self.svx][1] = self.axes_range_sliders["sl_x_max"].val
        msv_range[self.svy][0] = self.axes_range_sliders["sl_y_min"].val
        msv_range[self.svy][1] = self.axes_range_sliders["sl_y_max"].val
        self.set_mesh_grid()
        self.calc_phase_plane()
        self.update_phase_plane()

    def update_phase_plane(self):
        """ Clear the axes and redraw the phase-plane. """
        self.pp_ax.clear()
        self.pp_splt.clear()
        self.pp_splt.set_prop_cycle('color', get_color(self.model.nvar))
        self.pp_splt.plot(numpy.arange(TRAJ_STEPS + 1) * self.integrator.dt,
                          numpy.zeros((TRAJ_STEPS + 1, self.model.nvar)))
        if hasattr(self.pp_splt, 'autoscale'):
            self.pp_splt.autoscale(enable=True, axis='y', tight=True)
        self.pp_splt.legend(self.model.state_variables)
        self.plot_phase_plane()

    def update_state_variables(self, val):
        """
        Update the default state-variable values, used for non-visualised state
        variables, based of the current slider values.
        """
        for sv in self.sv_sliders:
            k = self.model.state_variables.index(sv)
            self.default_sv[k] = self.sv_sliders[sv].val

        self.calc_phase_plane()
        self.update_phase_plane()

    def set_mesh_grid(self):
        """
        Generate the phase-plane gridding based on currently selected
        state-variables and their range values.
        """
        xlo = self.model.state_variable_range[self.svx][0]
        xhi = self.model.state_variable_range[self.svx][1]
        ylo = self.model.state_variable_range[self.svy][0]
        yhi = self.model.state_variable_range[self.svy][1]

        self.X = numpy.mgrid[xlo:xhi:(NUMBEROFGRIDPOINTS * 1j)]
        self.Y = numpy.mgrid[ylo:yhi:(NUMBEROFGRIDPOINTS * 1j)]

    def set_state_vector(self):
        """
        Set up a vector containing the default state-variable values and create
        a filler(all zeros) for the coupling arg of the Model's dfun method.
        This method is called once at initialisation (show()).
        """
        # import pdb; pdb.set_trace()
        sv_mean = numpy.array([self.model.state_variable_range[key].mean() for key in self.model.state_variables])
        sv_mean = sv_mean.reshape((self.model.nvar, 1, 1))
        self.default_sv = sv_mean.repeat(self.model.number_of_modes, axis=2)
        self.no_coupling = numpy.zeros((self.model.nvar, 1,
                                        self.model.number_of_modes))

    def calc_phase_plane(self):
        """ Calculate the vector field. """
        svx_ind = self.model.state_variables.index(self.svx)
        svy_ind = self.model.state_variables.index(self.svy)

        # Calculate the vector field discretely sampled at a grid of points
        grid_point = self.default_sv.copy()
        self.U = numpy.zeros((NUMBEROFGRIDPOINTS, NUMBEROFGRIDPOINTS,
                              self.model.number_of_modes))
        self.V = numpy.zeros((NUMBEROFGRIDPOINTS, NUMBEROFGRIDPOINTS,
                              self.model.number_of_modes))
        for ii in range(NUMBEROFGRIDPOINTS):
            grid_point[svy_ind] = self.Y[ii]
            for jj in range(NUMBEROFGRIDPOINTS):
                # import pdb; pdb.set_trace()
                grid_point[svx_ind] = self.X[jj]

                d = self.model.dfun(grid_point, self.no_coupling)

                for kk in range(self.model.number_of_modes):
                    self.U[ii, jj, kk] = d[svx_ind, 0, kk]
                    self.V[ii, jj, kk] = d[svy_ind, 0, kk]

        # Colours for the vector field quivers
        # self.UVmag = numpy.sqrt(self.U**2 + self.V**2)

        # import pdb; pdb.set_trace()
        if numpy.isnan(self.U).any() or numpy.isnan(self.V).any():
            LOG.error("NaN")

    def plot_phase_plane(self):
        """ Plot the vector field and its nullclines. """
        # Set title and axis labels
        model_name = self.model.__class__.__name__
        self.pp_ax.set(title=model_name + " mode " + str(self.mode))
        self.pp_ax.set(xlabel="State Variable " + self.svx)
        self.pp_ax.set(ylabel="State Variable " + self.svy)

        # import pdb; pdb.set_trace()
        # Plot a discrete representation of the vector field
        if numpy.all(self.U[:, :, self.mode] + self.V[:, :, self.mode] == 0):
            self.pp_ax.set(title=model_name + " mode " + str(self.mode) + ": NO MOTION IN THIS PLANE")
            X, Y = numpy.meshgrid(self.X, self.Y)
            self.pp_quivers = self.pp_ax.scatter(X, Y, s=8, marker=".", c="k")
        else:
            self.pp_quivers = self.pp_ax.quiver(self.X, self.Y,
                                                self.U[:, :, self.mode],
                                                self.V[:, :, self.mode],
                                                # self.UVmag[:, :, self.mode],
                                                width=0.001, headwidth=8)

        # Plot the nullclines
        self.nullcline_x = self.pp_ax.contour(self.X, self.Y,
                                              self.U[:, :, self.mode],
                                              [0], colors="r")
        self.nullcline_y = self.pp_ax.contour(self.X, self.Y,
                                              self.V[:, :, self.mode],
                                              [0], colors="g")
        plt.draw()

    def plot_trajectory(self, x, y):
        """
        Plot a sample trajectory, starting at the position x,y in the
        phase-plane. This method is called as a result of a mouse click on the
        phase-plane.
        """
        svx_ind = self.model.state_variables.index(self.svx)
        svy_ind = self.model.state_variables.index(self.svy)

        # Calculate an example trajectory
        state = self.default_sv.copy()
        self.integrator.clamped_state_variable_indices = numpy.setdiff1d(
            numpy.r_[:len(self.model.state_variables)], numpy.r_[svx_ind, svy_ind])
        self.integrator.clamped_state_variable_values = self.default_sv[self.integrator.clamped_state_variable_indices]
        state[svx_ind] = x
        state[svy_ind] = y
        scheme = self.integrator.scheme
        traj = numpy.zeros((TRAJ_STEPS + 1, self.model.nvar, 1,
                            self.model.number_of_modes))
        traj[0, :] = state
        for step in range(TRAJ_STEPS):
            # import pdb; pdb.set_trace()
            state = scheme(state, self.model.dfun, self.no_coupling, 0.0, 0.0)
            traj[step + 1, :] = state

        self.pp_ax.scatter(x, y, s=42, c='g', marker='o', edgecolor=None)
        self.pp_ax.plot(traj[:, svx_ind, 0, self.mode],
                        traj[:, svy_ind, 0, self.mode])

        # Plot the selected state variable trajectories as a function of time
        self.pp_splt.plot(numpy.arange(TRAJ_STEPS + 1) * self.integrator.dt,
                          traj[:, :, 0, self.mode])

        plt.draw()

    def click_trajectory(self, event):
        """
        This method captures mouse clicks on the phase-plane and then uses the
        plot_trajectory() method to generate a sample trajectory.
        """
        if event.inaxes is self.pp_ax:
            x, y = event.xdata, event.ydata
            LOG.info('trajectory starting at (%f, %f)', x, y)
            self.plot_trajectory(x, y)


def _list_of_models():
    base = models_module.Model
    for key in dir(models_module):
        attr = getattr(models_module, key)
        if isinstance(attr, type) and issubclass(attr, base):
            if attr is base:
                continue
            first_para = attr.__doc__.replace('\n', ' ').replace('\t', ' ')[:100] + ' ...'
            yield attr.__name__, attr._ui_name


if __name__ == "__main__":

    import sys

    try:
        Model = getattr(models_module, sys.argv[1])
    except Exception:
        print("""
usage: python -m tvb.simulator.plot.phase_plane_interactive name_of_model

where name_of_model is one of

%s
        """ % (
            '\n'.join(map('{0[0]:>25} - {0[1]}'.format, _list_of_models()))
        ))
        sys.exit(1)

    ppi_fig = PhasePlaneInteractive(model=Model())
    ppi_fig.show()



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
This module is an example of how to validate model parameters and models.
The idea is to pick some model parameters then run a short simulation and see if it fails.
If it failed then you know that the model is broken with those settings.
You will use this module to run high dimensional parameter space explorations and get a report of what worked and what not.

.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
"""

import itertools
import numpy
from tvb.simulator.lab import *


# maybe a good idea given the high dimensionality of the model parameter space
# def sample_parameter_space_monte_carlo(some_distribution):
#     pass

def sample_parameter_space_cartesian(path_assignments):
    """
    Samples the parameter space systematically based on explicit values for each dimension (parameter).
    :param path_assignments: a list of (param_name, [param_values])
    >>> list(sample_parameter_space_cartesian([('a', [1, 2]), ('b', [3, 4])]))
    [{'a': 1, 'b': 3}, {'a': 1, 'b': 4}, {'a': 2, 'b': 3}, {'a': 2, 'b': 4}]
    """
    # 'transpose' path_assignments
    paths, values = list(zip(*path_assignments))
    # cartesian product of value assignments
    for value_assignment in itertools.product(*values):
        yield dict(list(zip(paths, value_assignment)))

def _set_sim_values(sim, path_assignment):
    """
    :param sim: A Simulator instance that will be modified
    :param path_assignment: a dict Ex {'model.a': 12, 'param.subpar.s': 100}
    The string_accessor should access a field on sim. Ex: model.param.subparam
    """
    for pth, val in path_assignment.items():
        code = 'sim.%s = val' % pth
        exec (code, {'sim': sim, 'val': val})


def run_exploration(sim, simulation_length, parameters):
    """
    Runs a series of simulations and reports weather they completed successfully or failed with FloatingPointError
    :param sim: a Simulator()
    :param simulation_length:
    :param parameters: [ {'model.a': 2, 'prop.subprop.w: 3 ...} ... ]
    :return:
    """

    print('starting model parameter space exploration')
    print('------------------------------------------')

    numpy.seterr(divide='raise', invalid='raise')

    for params in parameters:
        _set_sim_values(sim, params)
        sim.configure()
        print(params, end=' ')

        try:
            for traw in sim(simulation_length=simulation_length):
                traw = traw[0]
                state = traw[1]
                if not numpy.all(numpy.isfinite(state)):
                    raise FloatingPointError('infinities generated outside numpy')
            print('ok')
        except FloatingPointError:
            print('bad')


def example():
    model = models.Generic2dOscillator()
    white_matter = connectivity.Connectivity(load_default=True)
    det_int = integrators.HeunDeterministic(dt=2 ** -4)

    sim = simulator.Simulator(
        model=model,
        connectivity=white_matter,
        coupling=coupling.Linear(a=0.0126),
        integrator=det_int,
        monitors=[monitors.Raw()],
    )

    exploration_settings = [
        ('model.tau', [0.0, 1.0]),
        ('model.b', [-5.0, 15.0]),
        ('model.d', [0.0, 0.5, 1.0]),
        # ('coupling.a', [1,3])  # you can also use non-model parameter dimensions like this
    ]

    print('cartesian sampling settings: ')
    print(exploration_settings)
    print('------------------------------------------')

    run_exploration(
        sim,
        simulation_length=10,
        parameters=sample_parameter_space_cartesian(exploration_settings)
    )


if __name__ == '__main__':
    example()



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
This private module implements concrete declarative attributes
"""
import inspect
import numpy
import types
import typing
from collections.abc import Sequence
from ._declarative_base import _Attr, MetaType
from .ex import TraitValueError, TraitTypeError, TraitAttributeError, TraitFinalAttributeError
from tvb.basic.logger.builder import get_logger

if typing.TYPE_CHECKING:
    from ._core import HasTraits

# a logger for the whole traits system
log = get_logger('tvb.traits')


class Attr(_Attr):
    """
    An Attr declares the following about the attribute it describes:
    * the type
    * a default value shared by all instances
    * if the value might be missing
    * documentation
    It will resolve to attributes on the instance.
    """

    # This class is a python data descriptor.
    # For an introduction see https://docs.python.org/2/howto/descriptor.html

    def __init__(
            self, field_type, default=None, doc='', label='', required=True, final=False, choices=None
    ):
        # type: (type, typing.Any, str, str, bool, bool, typing.Optional[tuple]) -> None
        """
        :param field_type: the python type of this attribute
        :param default: A shared default value. Behaves like class level attribute assignment.
                        Take care with mutable defaults.
        :param doc: Documentation for this field.
        :param label: A short description.
        :param required: required fields should not be None.
        :param final: Final fields can only be assigned once.
        :param choices: A tuple of the values that this field is allowed to take.
        """
        super(Attr, self).__init__()
        self.field_type = field_type
        self.default = default
        self.doc = doc
        self.label = label
        self.required = bool(required)
        self.final = bool(final)
        self.choices = choices

    def __validate(self, value):
        """ check field_type and choices """
        if not isinstance(value, self.field_type) and not (
                inspect.isclass(self.default) and issubclass(value, self.field_type)):
            raise TraitTypeError("Attribute can't be set to an instance of {}".format(type(value)), attr=self)
        if self.choices is not None:
            if value not in self.choices and not (value is None and not self.required):
                raise TraitValueError("Value {!r} must be one of {}".format(value, self.choices), attr=self)

    # subclass api

    def _post_bind_validate(self):
        # type: () -> None
        """
        Validates this instance of Attr.
        This is called just after field_name is set, by MetaType.
        We do checks here and not in init in order to give better error messages.
        Attr should be considered initialized only after this has run
        """
        if not isinstance(self.field_type, type):
            msg = 'Field_type must be a type not {!r}. Did you mean to declare a default?'.format(
                self.field_type
            )
            raise TraitTypeError(msg, attr=self)

        skip_default_checks = self.default is None or isinstance(self.default, types.FunctionType)

        if not skip_default_checks:
            self.__validate(self.default)

        # heuristic check for mutability. might be costly. hasattr(__hash__) is fastest but less reliable
        try:
            hash(self.default)
        except TypeError:
            log.warning('Field seems mutable and has a default value. '
                        'Consider using a lambda as a value factory \n   attribute {}'.format(self))
        # we do not check here if we have a value for a required field
        # it is too early for that, owner.__init__ has not run yet

    def _validate_set(self, instance, value):
        # type: ('HasTraits', typing.Any) -> typing.Any
        """
        Called before updating the value of an attribute.
        It checks the type *AND* returns the valid value.
        You can override this for further checks. Call super to retain this check.
        Raise if checks fail.
        You should return a cleaned up value if validation passes
        """
        if value is None:
            if self.required:
                raise TraitValueError("Attribute is required. Can't set to None", attr=self)
            else:
                return value

        self.__validate(value)
        return value

    # descriptor protocol
    def __get__(self, instance, owner):
        # type: (typing.Optional['HasTraits'], 'MetaType') -> typing.Any
        self._assert_have_field_name()
        if instance is None:
            # called from class, not an instance
            return self
        # data is stored on the instance in a field with the same name
        # If field is not on the instance yet, return the class level default
        # (this attr instance is a class field, so the default is for the class)
        # This is consistent with how class fields work before they are assigned and become instance bound
        if self.field_name not in instance.__dict__:
            if (self.field_type != types.FunctionType and isinstance(self.default, types.FunctionType)
                    or inspect.isclass(self.default) and issubclass(self.default, self.field_type)):
                default = self.default()
            else:
                default = self.default

            # Unless we store the default on the instance, this will keep returning self.default()
            # when the default is a function. So if the default is mutable, any changes to it are
            # lost as a new one is created every time.
            instance.__dict__[self.field_name] = default

        value = instance.__dict__[self.field_name]
        if self.required and value is None:
            raise TraitAttributeError('required attribute referenced before assignment. '
                                      'Use a default or assign a value before reading it', attr=self)
        return value

    def __set__(self, instance, value):
        # type: ('HasTraits', typing.Any) -> None
        self._assert_have_field_name()

        if self.final:
            # non-set to set final transition happens when instance stored value becomes not none
            # getattr will call __get__. We want that in order to allow the default to be set.
            # If __set__ is called before a __get__ then no defaults have been assigned.
            # subtlety: if the value of this final field is not set then __get__ will raise
            #           getattr with a default value swallows that exception and returns false
            present_value = getattr(instance, self.field_name, None)
            if present_value is not None:
                raise TraitFinalAttributeError("can't write final attribute")

        value = self._validate_set(instance, value)

        instance.__dict__[self.field_name] = value

    def _defined_on_str_helper(self):
        if self.owner is not None:
            return '{}.{}.{} = {}'.format(
                self.owner.__module__,
                self.owner.__name__,
                self.field_name,
                type(self).__name__
            )
        else:
            return '{}'.format(type(self).__name__)

    def __str__(self):
        return '{}(field_type={}, default={!r}, required={})'.format(
            self._defined_on_str_helper(), self.field_type, self.default, self.required
        )


class Final(Attr):
    """
    An attribute that can only be set once.
    If a default is provided it counts as a set, so it cannot be written to.
    Note that if the default is a mutable type, the value is shared with all instances
    of the owning class.
    We cannot enforce true constancy in python
    """

    def __init__(self, default=None, field_type=None, doc='', label=''):
        """
        :param default: The constant value
        """
        # it would be nice if we could turn the default immutable. But this is unreasonable work in python
        # maybe a deep copy?
        if default is not None:
            field_type = type(default)

        if default is None and field_type is None:
            raise ValueError('Either a default or a field_type is required')

        super(Final, self).__init__(
            field_type=field_type, default=default, doc=doc, label=label, required=True, final=True
        )


class List(Attr):
    """
    The attribute is a list of values.
    Choices and type are reinterpreted as applying not to the list but to the elements of it
    """

    def __init__(self, of=object, default=(), doc='', label='', final=False, choices=None):
        # type: (type, tuple, str, str, bool, typing.Optional[tuple]) -> None
        super(List, self).__init__(
            field_type=Sequence,
            default=default,
            doc=doc,
            label=label,
            required=True,
            final=final,
            choices=None,
        )
        self.element_type = of
        self.element_choices = choices

    def __validate_elements(self, value):
        """ check that all elements are of the declared type and one of the declared choices """
        for i, el in enumerate(value):
            if not isinstance(el, self.element_type):
                raise TraitTypeError("value[{}] can't be of type {}".format(i, type(el)), attr=self)

        if self.element_choices is not None:
            for i, el in enumerate(value):
                if el not in self.element_choices:
                    raise TraitValueError(
                        "value[{}]=={!r} must be one of {}".format(i, el, self.element_choices),
                        attr=self
                    )

    def _post_bind_validate(self):
        super(List, self)._post_bind_validate()
        # check that the default contains elements of the declared type
        self.__validate_elements(self.default)

    def _validate_set(self, instance, value):
        value = super(List, self)._validate_set(instance, value)
        if value is None:
            # value is optional and missing, nothing to do here
            return
        self.__validate_elements(value)
        return value

    # __get__ __set__ here only for typing purposes, for better ide checking and autocomplete

    def __get__(self, instance, owner):
        # type: (typing.Optional[HasTraits], MetaType) -> typing.Sequence
        return super(List, self).__get__(instance, owner)

    def __set__(self, instance, value):
        # type: (HasTraits, typing.Sequence) -> None
        super(List, self).__set__(instance, value)

    def __str__(self):
        return '{}(of={}, default={!r}, required={})'.format(
            self._defined_on_str_helper(), self.element_type, self.default, self.required
        )


class _Number(Attr):

    def __validate(self, value):
        """ value should be safely cast to field type and choices must be enforced """
        if not isinstance(value, (int, float, complex, numpy.number)):
            # we have to check that the value is numeric before the can_cast check
            # as can_cast works with dtype strings as well
            # can_cast('i8', 'i32')
            raise TraitTypeError("can't be set to {!r}. Need a number.".format(value), attr=self)
        if not numpy.can_cast(value, self.field_type, 'safe'):
            raise TraitTypeError("can't be set to {!r}. No safe cast.".format(value), attr=self)
        if self.choices is not None:
            if value not in self.choices:
                raise TraitValueError("value {!r} must be one of {}".format(value, self.choices), attr=self)

    def _post_bind_validate(self):
        if self.default is not None:
            self.__validate(self.default)

    def _validate_set(self, instance, value):
        if value is None:
            if self.required:
                raise TraitValueError("is required. Can't set to None", attr=self)
            else:
                return value

        self.__validate(value)
        return self.field_type(value)


class Int(_Number):
    """
    Declares an integer
    This is different from Attr(field_type=int).
    The former enforces int subtypes
    This allows all integer types, including numpy ones that can be safely cast to the declared type
    according to numpy rules
    """

    def __init__(
            self, field_type=int, default=0, doc='', label='', required=True, final=False, choices=None
    ):
        super(_Number, self).__init__(
            field_type=field_type,
            default=default,
            doc=doc,
            label=label,
            required=required,
            final=final,
            choices=choices,
        )

    def _post_bind_validate(self):
        if not issubclass(self.field_type, (int, numpy.integer)):
            msg = 'field_type must be a python int or a numpy.integer not {!r}.'.format(self.field_type)
            raise TraitTypeError(msg, attr=self)
        # super call after the field_type check above
        super(Int, self)._post_bind_validate()


class Float(_Number):
    """
    Declares a float.
    This is different from Attr(field_type=float).
    The former enforces float subtypes.
    This allows any type that can be safely cast to the declared float type
    according to numpy rules.

    Reading and writing this attribute is slower than a plain python attribute.
    In performance sensitive code you might want to use plain python attributes
    or even better local variables.
    """

    def __init__(
            self, field_type=float, default=0, doc='', label='', required=True, final=False, choices=None
    ):
        super(_Number, self).__init__(
            field_type=field_type,
            default=default,
            doc=doc,
            label=label,
            required=required,
            final=final,
            choices=choices,
        )

    def _post_bind_validate(self):
        if not issubclass(self.field_type, (float, numpy.floating)):
            msg = 'field_type must be a python float or a numpy.floating not {!r}.'.format(self.field_type)
            raise TraitTypeError(msg, attr=self)
        # super call after the field_type check above
        super(Float, self)._post_bind_validate()


class Dim(Final):
    """
    A symbol that defines a dimension in a numpy array shape.
    It can only be set once. It is an int.
    Dimensions have to be set before any NArrays that reference them are used.
    """

    any = object()  # sentinel

    def __init__(self, doc=''):
        super(Dim, self).__init__(field_type=int, doc=doc)


class NArray(Attr):
    """
    Declares a numpy array.
    dtype enforces the dtype. The default dtype is float64.
    An optional symbolic shape can be given, as a tuple of Dim attributes from the owning class.
    The shape will be enforced, but no broadcasting will be done.
    domain declares what values are allowed in this array.
    It can be any object that can be checked for membership
    Defaults are checked if they are in the declared domain.
    For performance reasons this does not happen on every attribute set.
    """

    def __init__(
            self,
            default=None,
            required=True,
            doc='',
            label='',
            dtype=numpy.float64,
            shape=None,
            dim_names=(),
            domain=None,
    ):
        # type: (numpy.ndarray, bool, str, str, typing.Union[numpy.dtype, type, str], typing.Optional[typing.Tuple[Dim, ...]], typing.Tuple[str, ...], typing.Container) -> None
        """
        :param dtype: The numpy datatype. Defaults to float64. This is checked by neotraits.
        :param shape: An optional symbolic shape, tuple of Dim's declared on the owning class
        :param dim_names: Optional names for the names of the dimensions
        :param domain: Any type that can be checked for membership like xrange.
                       Represents the expected domain of the values in the array.
        """

        if numpy.issubdtype(dtype, numpy.integer):
            dtype = numpy.int64
        self.dtype = numpy.dtype(dtype)

        super(NArray, self).__init__(
            field_type=numpy.ndarray, default=default, required=required, doc=doc, label=label
        )
        self.shape = shape
        self.domain = domain  # anything that supports 3.1 in domain
        self.dim_names = dim_names

        if self.shape is not None:  # we have a shape
            if self.dim_names:  # and dim_names
                # ensure that len(shape) == len(dim_names)
                if len(self.shape) != len(self.dim_names):
                    raise TraitValueError('shape contradicts dim_names', attr=self)

            # maybe a over zealous type check
            for d in self.shape:
                if d is not Dim.any and type(d) != Dim:
                    raise TraitValueError("shape elements must be Dim's not {}".format(type(d)), attr=self)
            self.ndim = len(self.shape)
        elif self.dim_names:  # no shape but dim_names
            self.ndim = len(self.dim_names)
        else:
            self.ndim = None

    def __validate(self, value):
        """ check that ndim's and dtypes match"""
        if self.ndim is not None and value.ndim != self.ndim:
            raise TraitValueError("can't be set to an array with ndim {}".format(value.ndim), attr=self)

        if not numpy.can_cast(value.dtype, self.dtype, 'safe'):
            raise TraitTypeError("can't be set to an array of dtype {}".format(value.dtype), attr=self)

    def _post_bind_validate(self):
        if self.default is None:
            return
        if not isinstance(self.default, numpy.ndarray):
            msg = 'default {} should be a numpy.ndarray'.format(self.default)
            raise TraitTypeError(msg, attr=self)

        self.__validate(self.default)

        # we make the default a read only array
        self.default.setflags(write=False)

        # check that the default array values are in the declared domain
        # this may be expensive
        if self.domain is not None and self.default is not None:
            for e in self.default.flat:
                if e not in self.domain:
                    msg = 'default contains values out of the declared domain. Ex {}'.format(e)
                    log.warning('{} \n   attribute  {}'.format(msg, self))

                    break

    def _lookup_expected_shape(self, instance):
        """ look up expected shape on the instance """
        expected_shape = []

        for dim_attr in self.shape:
            if dim_attr is Dim.any:
                expected_dim = Dim.any
            else:
                try:
                    # invoke Dim's __get__(instance)
                    expected_dim = getattr(instance, dim_attr.field_name)
                except TraitAttributeError:
                    # re-raise with a better error message
                    msg = "Narray's shape references undefined dimension <{}>. " \
                          "Set it before accessing this array"
                    raise TraitAttributeError(msg.format(dim_attr.field_name), attr=self)
            expected_shape.append(expected_dim)
        return expected_shape

    def _validate_set(self, instance, value):
        value = super(NArray, self)._validate_set(instance, value)
        if value is None:
            # value is optional and missing, nothing to do here
            return
        self.__validate(value)
        # we should know here the concrete shape
        # check it

        if self.shape is not None:
            expected_shape = self._lookup_expected_shape(instance)

            for expected_dim, value_dim in zip(expected_shape, value.shape):
                if expected_dim is Dim.any:
                    continue
                if value_dim != expected_dim:
                    raise TraitValueError(
                        'Shape mismatch. Expected {}. Given {}. Not broadcasting'.format(
                            expected_shape, value.shape
                        )
                    )

        return value.astype(self.dtype)

    # here only for typing purposes, so ide's can get better suggestions
    def __get__(self, instance, owner):
        # type: (typing.Optional['HasTraits'], 'MetaType') -> typing.Union[numpy.ndarray, 'NArray']
        return super(NArray, self).__get__(instance, owner)

    def __set__(self, instance, value):
        # type: (HasTraits, numpy.ndarray) -> None
        super(NArray, self).__set__(instance, value)

    def __str__(self):
        return '{}(label={!r}, dtype={}, default={!r}, dim_names={}, ndim={}, required={})'.format(
            self._defined_on_str_helper(),
            self.label,
            self.dtype,
            self.default,
            self.dim_names,
            self.ndim,
            self.required,
        )


class EnumAttr(Attr):
    def __init__(self, field_type=None, default=None, doc='', label='', required=True):
        """
        :param default: The default enum value
        """

        if field_type is None and default is not None:
            field_type = type(default)

        if default is None and field_type is None:
            raise ValueError('Either a default or a field_type is required')

        super(EnumAttr, self).__init__(
            field_type=field_type, default=default, doc=doc, label=label, required=required, choices=tuple(field_type)
        )

    def __validate(self, value):
        """
        value either has to be in the enum choices or its tpye has to be in the value list of enum choices
        """
        if value in self.choices:
            return

        if type(value) in [choice.value for choice in self.choices]:
            return

        raise TraitTypeError("Attribute can't be set to an instance of {}".format(type(value)), attr=self)

    def _validate_set(self, instance, value):
        # type: ('HasTraits', typing.Any) -> typing.Any
        """
        Same method as in Attr.
        The reason we override it is to not call the __validate method from the superclass.
        """
        if value is None:
            if self.required:
                raise TraitValueError("Attribute is required. Can't set to None", attr=self)
            else:
                return value

        self.__validate(value)
        return value

    def _post_bind_validate(self):
        if self.default is not None:
            self.__validate(self.default)

    # def __get__(self, instance, owner):
    #     # type: (typing.Optional['TupleEnum'], 'MetaType') -> typing.Union[Enum, 'Enum']
    #     return super(EnumAttr).__get__(instance, owner)
    #
    # def __set__(self, instance, value):
    #     # type: (TupleEnum, Enum) -> None
    #     super(EnumAttr).__set__(instance, value)


class Range(object):
    """
    Defines a domain like the one that numpy.arange generates
    Points are precisely equidistant but the largest point is <= hi
    """

    def __init__(self, lo, hi, step=1.0):
        self.lo = lo
        self.hi = hi
        self.step = step if abs(hi - lo) > abs(step) else abs(hi - lo)

    def __contains__(self, item):
        """ true if item between lo and high. ignores the step"""
        return self.lo <= item < self.hi

    def to_array(self):
        return numpy.arange(self.lo, self.hi, self.step)

    def __repr__(self):
        return 'Range(lo={}, hi={}, step={})'.format(self.lo, self.hi, self.step)


class LinspaceRange(object):
    """
    Defines a domain with precise endpoints but the points are not precisely equidistant
    Similar to numpy.linspace
    """

    def __init__(self, lo, hi, npoints=50):
        self.lo = lo
        self.hi = hi
        self.npoints = npoints

    def __contains__(self, item):
        """ true if item between lo and high. ignores the step"""
        return self.lo <= item < self.hi

    def to_array(self):
        return numpy.linspace(self.lo, self.hi, self.npoints)

    def __repr__(self):
        return 'LinspaceRange(lo={}, hi={}, step={})'.format(self.lo, self.hi, self.npoints)



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#
import numpy
import numpy as np
import scipy.sparse as ss

from .backendtestbase import BaseTestSim
from tvb.simulator.backend.nb_mpr import NbMPRBackend
from tvb.simulator.lab import *


class TestNbSim(BaseTestSim):

    #def get_run_sim_chunked(self, print_source=False):
    #    template = '<%include file="nb-montbrio.py.mako"/>'
    #    content = dict(foo='bar')

    #    return NbMPRBackend().build_py_func(template, content, name='run_sim_tavg_chunked', print_source=print_source)

    def _random_network(self, N=300, density=0.1, low=5, high=250, speed=np.inf):
        weights = ss.random(N,N, density=density, format='lil')
        weights.setdiag(0)
        weights = weights.tocsr()
        lengths = weights.copy()
        lengths.data[:] = np.random.uniform(low=low, high=high, size=lengths.data.shape[0])

        N = weights.shape[0]
        conn = connectivity.Connectivity(
                weights=weights.toarray(),
                tract_lengths=lengths.toarray(),
                region_labels=np.array( [f'roi_{i}' for i in range(N)]), 
                centres=np.zeros(N),
                speed=np.r_[speed]
        )
        return conn

    def test_local_deterministic(self):
        dt = 0.01
        G = 0.

        sim = simulator.Simulator(
            model=models.MontbrioPazoRoxin(),
            coupling=coupling.Linear(a=np.array([G])),
            connectivity=self._random_network(),
            conduction_speed=np.inf,
            monitors=[
                monitors.Raw()
            ],
            integrator=integrators.HeunStochastic( # matching integrator
                dt=dt, 
                noise=noise.Additive(
                    nsig=np.array([0.0, 0.0]),
                    noise_seed=42
                )
            )
        ).configure()

        (pdq_t, pdq_d), = NbMPRBackend().run_sim(sim, nstep=1)
        (raw_t, raw_d), = sim.run(simulation_length=1)

        np.testing.assert_allclose(raw_d[0,:], pdq_d[0,:], rtol=1e-5)

    def test_local_deterministic_spatial(self):
        dt = 0.01
        G = 0.
        N=300

        sim = simulator.Simulator(
            model=models.MontbrioPazoRoxin(
                eta=np.random.uniform(
                    models.MontbrioPazoRoxin.eta.domain.lo,
                    models.MontbrioPazoRoxin.eta.domain.hi,
                    size=N
                )
            ),
            coupling=coupling.Linear(a=np.array([G])),
            connectivity=self._random_network(N),
            conduction_speed=np.inf,
            monitors=[
                monitors.Raw()
            ],
            integrator=integrators.HeunStochastic( # matching integrator
                dt=dt, 
                noise=noise.Additive(
                    nsig=np.array([0.0, 0.0]),
                    noise_seed=42
                )
            )
        ).configure()

        (pdq_t, pdq_d), = NbMPRBackend().run_sim(sim, nstep=1)
        (raw_t, raw_d), = sim.run(simulation_length=1)

        np.testing.assert_allclose(raw_d[0,:], pdq_d[0,:], rtol=1e-5, atol=1e-3)


    def test_local_stochastic(self):
        #run_sim = self._get_run_sim(print_source=True)
        dt = 0.01
        G = 0.

        sim = simulator.Simulator(
            model=models.MontbrioPazoRoxin(),
            coupling=coupling.Linear(a=np.array([G])),
            connectivity=self._random_network(),
            conduction_speed=np.inf,
            monitors=[
                monitors.Raw()
            ],
            integrator=integrators.HeunStochastic( 
                dt=dt, 
                noise=noise.Additive(
                    nsig=np.array([0.01, 0.02]),
                    noise_seed=42
                )
            )
        ).configure()

        sim_det = simulator.Simulator(
            model=models.MontbrioPazoRoxin(),
            coupling=coupling.Linear(a=np.array([G])),
            connectivity=self._random_network(),
            conduction_speed=np.inf,
            monitors=[
                monitors.Raw()
            ],
            integrator=integrators.HeunDeterministic( 
                dt=dt, 
            )
        ).configure()

        (pdq_t, pdq_d), = NbMPRBackend().run_sim(sim, nstep=200)

        (raw_t, raw_d), = sim.run(simulation_length=1)
        (raw_t_det, raw_d_det), = sim_det.run(simulation_length=1)
        r_tvb, V_tvb = raw_d[0,:,:,0]
        r_tvb_det, V_tvb_det = raw_d_det[0,:,:,0]
        r_pdq, V_pdq = pdq_d[0,:,:,0] 

        np.testing.assert_allclose(
                np.mean(r_tvb_det - r_tvb),
                np.mean(r_tvb_det - r_pdq),
                atol=1e-2 # zero mean doesn't play well with rtol
        )
        np.testing.assert_allclose(
                np.mean(V_tvb_det - V_tvb),
                np.mean(V_tvb_det - V_pdq),
                atol=1e-2
        )
        np.testing.assert_allclose(
                np.std(r_tvb_det - r_tvb), 
                np.std(r_tvb_det - r_pdq),
                rtol=1e-2
        )
        np.testing.assert_allclose(
                np.std(V_tvb_det - V_tvb),
                np.std(V_tvb_det - V_pdq),
                rtol=1e-2
        )

    def test_network_deterministic_nodelay(self):
        dt = 0.01
        G = 0.8

        sim = simulator.Simulator(
            model=models.MontbrioPazoRoxin(),
            coupling=coupling.Linear(a=np.array([G])),
            connectivity=self._random_network(),
            conduction_speed=np.inf,
            monitors=[
                monitors.Raw()
            ],
            integrator=integrators.HeunStochastic( 
                dt=dt, 
                noise=noise.Additive(
                    nsig=np.array([0.0, 0.0]),
                    noise_seed=42
                )
            )
        ).configure()

        with numpy.errstate(all='ignore'):
            (pdq_t, pdq_d), = NbMPRBackend().run_sim(sim, nstep=1, compatibility_mode=True)
            r_pdq, V_pdq = pdq_d[0,:,:,0] 

            (raw_t, raw_d), = sim.run(simulation_length=1)
            r_tvb, V_tvb = raw_d[0, :, :, 0]

        np.testing.assert_allclose(r_tvb, r_pdq, rtol=1e-4, atol=1e-3)
        np.testing.assert_allclose(V_tvb, V_pdq, rtol=1e-4, atol=1e-3)
        
    def test_network_deterministic_delay(self):
        dt = 0.01
        G = 0.8
        speed=2.

        sim = simulator.Simulator(
            model=models.MontbrioPazoRoxin(),
            coupling=coupling.Linear(a=np.array([G])),
            connectivity=self._random_network(speed=speed),
            conduction_speed=speed,
            monitors=[
                monitors.Raw()
            ],
            integrator=integrators.HeunStochastic( 
                dt=dt, 
                noise=noise.Additive(
                    nsig=np.array([0.0, 0.0]),
                    noise_seed=42
                )
            )
        ).configure()

        (pdq_t, pdq_d), = NbMPRBackend().run_sim(sim, nstep=1, compatibility_mode=True)
        r_pdq, V_pdq = pdq_d[0,:,:,0] 

        (raw_t, raw_d), = sim.run(simulation_length=1)
        r_tvb, V_tvb = raw_d[0,:,:,0]

        np.testing.assert_allclose(r_tvb, r_pdq, rtol=1e-4)
        np.testing.assert_allclose(V_tvb, V_pdq, rtol=1e-4, atol=1e-3)

    def test_tavg_chunking(self):
        dt = 0.01
        G = 0.1
        speed=2.

        sim = simulator.Simulator(
            model=models.MontbrioPazoRoxin(),
            coupling=coupling.Linear(a=np.array([G])),
            connectivity=self._random_network(speed=speed),
            conduction_speed=speed,
            monitors=[
                monitors.TemporalAverage(period=1)
            ],
            integrator=integrators.HeunStochastic( 
                dt=dt, 
                noise=noise.Additive(
                    nsig=np.array([0.0, 0.0]),
                    noise_seed=42
                )
            )
        ).configure()

        (pdq_t, pdq_d), = NbMPRBackend().run_sim(sim, nstep=2000, chunksize=2000, compatibility_mode=True)
        r_pdq = pdq_d[:,0,:,0]
        V_pdq = pdq_d[:,1,:,0]
        (pdq_chu_t, pdq_chu_d), = NbMPRBackend().run_sim(sim, nstep=2000, chunksize=500, compatibility_mode=True)
        r_pdq_chu = pdq_d[:,0,:,0]
        V_pdq_chu = pdq_d[:,1,:,0]

        (raw_t, raw_d), = sim.run(simulation_length=20)
        r_tvb = raw_d[:,0,:,0]
        V_tvb = raw_d[:,1,:,0]

        np.testing.assert_allclose(pdq_t,raw_t)
        np.testing.assert_allclose(pdq_chu_t,raw_t)

        np.testing.assert_allclose(r_pdq, r_tvb, atol=1e-3, rtol=1e-3)
        np.testing.assert_allclose(V_pdq, V_tvb, atol=1e-3, rtol=1e-3)
        # think a bit about the tolerances...  TVB stores in floats, so that 
        # can accumulate. Might be a good idea to test agains history with 
        # double typed buffer.
        np.testing.assert_allclose(r_pdq_chu, r_tvb, atol=1e-3, rtol=1e-3)
        np.testing.assert_allclose(V_pdq_chu, V_tvb, atol=1e-3, rtol=1e-3)

    def test_stim(self):

        G = 0.525
        nsigma = 0.0
        conn_speed=2.

        conn = connectivity.Connectivity()
        conn.motif_all_to_all(number_of_regions=3)
        conn.speed = np.r_[conn_speed]
        conn.weights[:] = 0.
        conn.weights[0,1] = 1.
        conn.weights[1,2] = 1.
        conn.weights[2,0] = 1.
        conn.centres_spherical(number_of_regions=conn.number_of_regions)
        conn.create_region_labels(mode='alphabetic')

        conn.weights = conn.weights/np.max(conn.weights)
        np.fill_diagonal(conn.weights, 0.)
        conn.configure()

        weighting = np.zeros((conn.number_of_regions, ))
        weighting[[1]] = 5.0

        eqn_t = equations.PulseTrain(
            parameters={
                'onset': 50,
                'T': 10000.0 ,
                'tau': 10.0,
                'amp': 1.
            }
        )

        stimulus = patterns.StimuliRegion(
            temporal=eqn_t,
            connectivity=conn,
            weight=weighting)


        simulation_length = 150
        seed=42

        ic = np.zeros( (1, 2, 3, 1))
        ic[:,1,:,:] = -2.

        sim = simulator.Simulator(
            model=models.MontbrioPazoRoxin(
                eta   = np.r_[-5.0],
                J     = np.r_[15.],
                Delta = np.r_[1.],
            ),
            connectivity=conn,
            coupling=coupling.Linear(
                a=np.array([G])
            ),
            conduction_speed=conn_speed,
            integrator=integrators.HeunStochastic(
                dt=0.01,
                noise=noise.Additive(
                    nsig=np.array(
                        [nsigma,nsigma*2]
                    ),
                    noise_seed=seed)
            ),
            monitors=[
                monitors.Raw()
            ],
            initial_conditions=ic,
            stimulus=stimulus,
            simulation_length=simulation_length
        ).configure()

        (pdq_t, pdq_d), = NbMPRBackend().run_sim(sim, simulation_length=simulation_length, compatibility_mode=True)
        (raw_t, raw_d), = sim.run(simulation_length=simulation_length)

        np.testing.assert_allclose(raw_d, pdq_d, rtol=1e-4)


    def test_stim_chunked(self):
        G = 0.525
        nsigma = 0.0
        conn_speed=2.
        transient = 0

        conn = connectivity.Connectivity()
        conn.motif_all_to_all(number_of_regions=3)
        conn.speed = np.r_[conn_speed]
        conn.weights[:] = 0.
        conn.weights[0,1] = 1.
        conn.weights[1,2] = 1.
        conn.weights[2,0] = 1.
        conn.centres_spherical(number_of_regions=conn.number_of_regions)
        conn.create_region_labels(mode='alphabetic')

        conn.weights = conn.weights/np.max(conn.weights)
        np.fill_diagonal(conn.weights, 0.)
        conn.configure()

        weighting = np.zeros((conn.number_of_regions, ))
        weighting[[1]] = 5.0

        eqn_t = equations.PulseTrain(
            parameters={
                'onset': 50,
                'T': 10000.0 ,
                'tau': 10.0,
                'amp': 1.
            }
        )

        stimulus = patterns.StimuliRegion(
            temporal=eqn_t,
            connectivity=conn,
            weight=weighting)


        simulation_length = 150
        seed=42

        ic = np.zeros( (1, 2, 3, 1))
        ic[:,1,:,:] = -2.

        sim = simulator.Simulator(
            model=models.MontbrioPazoRoxin(
                eta   = np.r_[-5.0],
                J     = np.r_[15.],
                Delta = np.r_[1.],
            ),
            connectivity=conn,
            coupling=coupling.Linear(
                a=np.array([G])
            ),
            conduction_speed=conn_speed,
            integrator=integrators.HeunStochastic(
                dt=0.01,
                noise=noise.Additive(
                    nsig=np.array(
                        [nsigma,nsigma*2]
                    ),
                    noise_seed=seed)
            ),
            monitors=[
                monitors.TemporalAverage(period=1.0)
            ],
            initial_conditions=ic,
            stimulus=stimulus,
            simulation_length=simulation_length
        ).configure()


        (pdq_t, pdq_d), = NbMPRBackend().run_sim(sim, simulation_length=simulation_length, compatibility_mode=True, chunksize=5000)

        (tvb_t, tvb_d), = sim.run(simulation_length=simulation_length)

        np.testing.assert_allclose(tvb_d, pdq_d, rtol=1e-4)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#

"""
Data descriptors for declaring workspace for algorithms and checking usage.

.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>

"""

import numpy
import collections
import weakref
import six
from .common import get_logger

LOG = get_logger(__name__)

# StaticAttr prevents descriptors from placing instance-owned, descriptor storage

class StaticAttr(object):
    "Base class which requires all attributes to be declared at class level."

    def __setattr__(self, name, value):
        attr_exists = hasattr(self, name)
        if not hasattr(self, name):
            raise AttributeError('%r has no attr %r.' % (self, name))
        else:
            super(StaticAttr, self).__setattr__(name, value)


class ImmutableAttrError(AttributeError):
    "Error due to modifying an immutable attribute."
    pass


class IncorrectTypeAttrError(AttributeError):
    "Error due to using incorrect type to set attribute value."
    pass


class NDArray(StaticAttr):
    "Data descriptor for a NumPy array, with type, mutability and shape checking."

    # Owner can provide array constructor via _array_ctor attr, i.e. NumPy or PyOpenCL, etc. ?

    State = collections.namedtuple('State', 'array initialized')

    shape, dtype, read_only, instance_state = (), None, True, {}

    def __init__(self, shape, dtype, read_only=True):
        self.shape = shape # may have strings which eval in owner ns
        self.dtype = dtype
        self.read_only = read_only
        self.instance_state = weakref.WeakKeyDictionary()

    def _make_array(self, instance):
        shape = []
        for dim in self.shape:
            if isinstance(dim, str):
                dim = getattr(instance, dim)
            elif isinstance(dim, Dim):
                if instance in dim.instance_state:
                    dim = dim.instance_state[instance].value
                else:
                    raise AttributeError('Dimension referenced before definition.')
            else:
                raise TypeError('expect int, str but found %r' % (type(dim), ))
            shape.append(dim)
        array = numpy.empty(shape, self.dtype)
        if self.read_only and hasattr(array, 'setflags'):
            array.setflags(write=False)
        return array

    def _get_or_create_state(self, instance):
        if instance not in self.instance_state:
            array = self._make_array(instance)
            self.instance_state[instance] = NDArray.State(array, False)
        return self.instance_state[instance]

    def __get__(self, instance, _):
        if instance is None:
            LOG.debug('NDArray returning self for None instance.')
            return self
        else:
            return self._get_or_create_state(instance).array

    def __set__(self, instance, value):
        state = self._get_or_create_state(instance)
        if self.read_only:
            if state.initialized:
                raise ImmutableAttrError('Cannot modify an immutable ndarray.')
            else:
                state.array.setflags(write=True)
        # set with [:] to ensure shape compat and safe type coercion
        _, value = numpy.broadcast_arrays(state.array, value)
        state.array[:] = value
        if self.read_only:
            state.array.setflags(write=False)
        if not state.initialized:
            self.instance_state[instance] = NDArray.State(state.array, True)


class Final(object):
    "A descriptor for an attribute, possibly type-checked, that once initialized, cannot be changed."

    State = collections.namedtuple('State', 'value initialized')

    def __init__(self, type=None):
        self.instance_state = weakref.WeakKeyDictionary()
        self.type = type

    def _get_or_create_state(self, instance):
        if instance not in self.instance_state:
            self.instance_state[instance] = Final.State(None, False)
        return self.instance_state[instance]

    def _correct_type(self, value):
        return isinstance(value, self.type)

    def __set__(self, instance, value):
        state = self._get_or_create_state(instance) # type: Final.State
        if state.initialized:
            raise AttributeError('final attribute cannot be set.')
        else:
            if self.type and not self._correct_type(value):
                raise AttributeError('value %r does not match expected type %r'
                                      % (value, self.type))
            self.instance_state[instance] = Final.State(value, True)

    def __get__(self, instance, owner):
        if instance is None:
            LOG.debug('Final returning self for None instance.')
            return self
        else:
            return self._get_or_create_state(instance).value


class Dim(Final):
    "Specialization of Final to int/long type."

    def __init__(self):
        super(Dim, self).__init__(int)

    def _correct_type(self, value):
        return isinstance(value, six.integer_types) \
               or numpy.issubdtype(type(value), numpy.integer)


# TODO
class Workspace(StaticAttr):
    pass


# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#
"""
Created on Mar 20, 2013

.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

import numpy
from tvb.tests.library.base_testcase import BaseTestCase
from tvb.datatypes import volumes


class TestVolumes(BaseTestCase):
    """
    Tests the defaults for `tvb.datatypes.volumes` module.
    """

    def test_volume(self):
        dt = volumes.Volume(origin=numpy.array([]), voxel_size=numpy.array([]))
        summary_info = dt.summary_info()
        assert summary_info['Origin'] is not None
        assert summary_info['Voxel size'] is not None
        assert summary_info['Volume type'] == 'Volume'
        assert summary_info['Units'] == 'mm'
        assert dt.origin is not None
        assert dt.voxel_size is not None
        assert dt.voxel_unit == 'mm'



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

from tvb.tests.library.base_testcase import BaseTestCase
from tvb.basic.neotraits._attr import Range


class TestRange(BaseTestCase):

    def assert_with_epsilon_error(self, list_one, list_two):
        epsilon = 0.000001
        value_for_return = True
        if len(list_one) != len(list_two):
            return False
        for i in range(0, len(list_one)):
            if (list_one[i] + epsilon) <= list_two[i] <= (list_one[i] + epsilon):
                continue
            elif (list_one[i] - epsilon) <= list_two[i] <= (list_one[i] - epsilon):
                continue
            elif (list_one[i] - epsilon) <= list_two[i] <= (list_one[i] + epsilon):
                continue
            elif (list_one[i] + epsilon) <= list_two[i] <= (list_one[i] - epsilon):
                continue
            else:
                value_for_return = False
        return value_for_return

    def test_generates_range_with_start_and_stop_provided(self):
        floats = list(Range(lo=0.0, hi=3.0, step=1.0).to_array())
        assert TestRange.assert_with_epsilon_error(self, floats, list([0.0, 1.0, 2.0]))

    def test_generates_range_with_another_start_stop_and_step(self):
        floats = list(Range(lo=1.0, hi=8.0, step=2.0).to_array())
        assert TestRange.assert_with_epsilon_error(self, floats, list([1.0, 3.0, 5.0, 7.0]))

    def test_generates_range_with_start_stop_and_step_smaller_than_one(self):
        floats = list(Range(lo=0.0, hi=0.5, step=0.1).to_array())
        assert TestRange.assert_with_epsilon_error(self, floats, list([0.0, 0.1, 0.2, 0.3, 0.4]))

    def test_generates_range_with_start_stop_and_periodic_repeating_step(self):
        floats = list(Range(lo=0.0, hi=1.1, step=1. / 3).to_array())
        assert TestRange.assert_with_epsilon_error(self, floats,
                                                   list([0.0, 0.333333333333, 0.666666666666, 0.999999999999]))

    def test_generates_range_with_negative_end(self):
        floats = list(Range(lo=1.0, hi=-3.0, step=-1.0).to_array())
        assert TestRange.assert_with_epsilon_error(self, floats, list([1.0, 0.0, -1.0, -2.0]))

    def test_generates_range_with_negative_end_including_both(self):
        floats = list(Range(lo=-3.0, hi=1.1, step=1.0).to_array())
        assert TestRange.assert_with_epsilon_error(self, floats, list([-3.0, -2.0, -1.0, 0.0, 1.0]))

    def test_generates_range_with_challenging_float_point_arithmetics(self):
        floats = list(Range(lo=0.0, hi=2.2, step=0.7).to_array())
        assert TestRange.assert_with_epsilon_error(self, floats, list([0.0, 0.7, 1.4, 2.1]))



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
Filler analyzer: Takes a TimeSeries object and returns two Floats.


These metrics are described and used in:

Hellyer et al. The Control of Global Brain Dynamics: Opposing Actions
of Frontoparietal Control and Default Mode Networks on Attention. 
The Journal of Neuroscience, January 8, 2014,  34(2):451– 461

Proxy of spatial coherence (V): 

Proxy metastability (M): the variability in spatial coherence of the signal
globally or locally (within a network) over time.

Proxy synchrony (S) : the reciprocal of mean spatial variance across time.

.. moduleauthor:: Paula Sanz Leon <paulala@tvb.invalid>

"""

import numpy
from tvb.basic.logger.builder import get_logger


def remove_mean(x, axis):
    """
    Remove mean from numpy array along axis
    """
    # Example for demean(x, 2) with x.shape == 2,3,4,5
    # m = x.mean(axis=2) collapses the 2'nd dimension making m and x incompatible
    # so we add it back m[:,:, np.newaxis, :]
    # Since the shape and axis are known only at runtime
    # Calculate the slicing dynamically
    return x - numpy.expand_dims(x.mean(axis=axis), axis)


r"""
Subtract the mean time-series and compute. 

Input:
TimeSeries DataType
    
Output: 
Float, Float
    
The two metrics given by this analyzers are a proxy for metastability and synchrony. 
The underlying dynamical model used in the article was the Kuramoto model.

.. math::
        V(t) &= \frac{1}{N} \sum_{i=1}^{N} |S_i(t) - <S(t)>| \\
        M(t) &= \sqrt{E[V(t)^{2}]-(E[V(t)])^{2}} \\
        S(t) &= \frac{1}{\bar{V(t)}}

"""

log = get_logger(__name__)


def compute_proxy_metastability_metric(params):
    """
    # type: dict(TimeSeries, float, int) -> (float, float)
    Compute the zero centered variance of node variances for the time_series.

    Parameters
    ----------
    params : a dictionary containing
        time_series : TimeSeries
        Input time series for which the metric will be computed.

        start_point : float
        Determines how many points of the TimeSeries will be discarded before computing the metric

        segment : int
        Divides the input time-series into discrete equally sized sequences and use the last segment to compute
        the metric. Only used when the start point is larger than the time-series length
    """

    time_series = params['time_series']
    start_point = params['start_point']
    segment = params['segment']
    shape = time_series.data.shape
    tpts = shape[0]

    if start_point != 0.0:
        start_tpt = start_point / time_series.sample_period
        log.debug("Will discard: %s time points" % start_tpt)
    else:
        start_tpt = 0

    if start_tpt > tpts:
        log.warning("The time-series is shorter than the starting point")
        log.debug("Will divide the time-series into %d segments." % segment)
        # Lazy strategy
        start_tpt = int((segment - 1) * (tpts // segment))

    start_tpt = int(start_tpt)
    time_series_diffs = remove_mean(time_series.data[start_tpt:, :], axis=2)
    v_data = abs(time_series_diffs).mean(axis=2)

    # handle state-variables & modes
    cat_tpts = v_data.shape[0] * shape[1] * shape[3]
    v_data = v_data.reshape((cat_tpts,), order="F")
    # std across time-points
    metastability = v_data.std(axis=0)
    synchrony = 1. / v_data.mean(axis=0)

    return {"Metastability": metastability,
            "Synchrony": synchrony}



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
Numba backend which uses templating to generate simulation
code.

.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>

"""

import os
import importlib
import numpy as np
import autopep8

from .templates import MakoUtilMix
from tvb.simulator.lab import *


class NbMPRBackend(MakoUtilMix):

    def build_py_func(self, template_source, content, name='kernel', print_source=False,
            modname=None):
        "Build and retrieve one or more Python functions from template."
        source = self.render_template(template_source, content)
        source = autopep8.fix_code(source)
        if print_source:
            print(self.insert_line_numbers(source))
        if modname is not None:
            return self.eval_module(source, name, modname)
        else:
            return self.eval_source(source, name)

    def eval_source(self, source, name):
        globals_ = {}
        try:
            exec(source, globals_)
        except Exception as exc:
            if not print_source:
                print(self._insert_line_numbers(source))
            raise exc
        fns = [globals_[n] for n in name.split(',')]
        return fns[0] if len(fns)==1 else fns

    def eval_module(self, source, name, modname):
        here = os.path.abspath(os.path.dirname(__file__))
        genp = os.path.join(here, 'templates', 'generated')
        with open(f'{genp}/{modname}.py', 'w') as fd:
            fd.write(source)
        fullmodname = f'tvb.simulator.backend.templates.generated.{modname}'
        mod = importlib.import_module(fullmodname)
        fns = [getattr(mod,n) for n in name.split(',')]
        return fns[0] if len(fns)==1 else fns

    def check_compatibility(self, sim): 
        def check_choices(val, choices):
            if not isinstance(val, choices):
                raise NotImplementedError("Unsupported simulator component. Given: {}\nExpected one of: {}".format(val, choices))
        # monitors
        if len(sim.monitors) > 1:
            raise NotImplementedError("Configure with one monitor.")
        check_choices(sim.monitors[0], (monitors.Raw, monitors.TemporalAverage))
        # integrators
        check_choices(sim.integrator, integrators.HeunStochastic)
        # models
        check_choices(sim.model, models.MontbrioPazoRoxin)
        # coupling
        check_choices(sim.coupling, coupling.Linear)
        # surface
        if sim.surface is not None:
            raise NotImplementedError("Surface simulation not supported.")
        # stimulus evaluated outside the backend, no restrictions


    def run_sim(self, sim, nstep=None, simulation_length=None, chunksize=100000, compatibility_mode=False, print_source=False):
        assert nstep is not None or simulation_length is not None or sim.simulation_length is not None

        self.check_compatibility(sim)
        if nstep is None:
            if simulation_length is None:
                simulation_length = sim.simulation_length
            nstep = int(np.ceil(simulation_length/sim.integrator.dt))

        if isinstance(sim.monitors[0], monitors.Raw):
            svar_bufs = self._run_sim_plain(sim, nstep, compatibility_mode=compatibility_mode, print_source=print_source)
            time = np.arange(svar_bufs[0].shape[1]) * sim.integrator.dt
        elif isinstance(sim.monitors[0], monitors.TemporalAverage):
            svar_bufs = self._run_sim_tavg_chunked(sim, nstep, chunksize=chunksize, compatibility_mode=compatibility_mode, print_source=print_source)
            T = sim.monitors[0].period
            time = np.arange(svar_bufs[0].shape[1]) * T + 0.5 * T
        else:
            raise NotImplementedError("Only Raw or TemporalAverage monitors supported.")
        data = np.concatenate(
                [svar_buf.T[:,np.newaxis,:, np.newaxis] for svar_buf in svar_bufs],
                axis=1
        )
        return (time, data),   

    def _run_sim_plain(self, sim, nstep=None, compatibility_mode=False, print_source=True):
        template = '<%include file="nb-montbrio.py.mako"/>'
        content = dict(
                compatibility_mode=compatibility_mode, 
                sim=sim
        ) 
        integrate = self.build_py_func(template, content, name='integrate', print_source=print_source)

        horizon = sim.connectivity.horizon
        buf_len = horizon + nstep
        N = sim.connectivity.number_of_regions
        gf = sim.integrator.noise.gfun(None)

        svar_bufs = [buf for buf in sim.integrator.noise.generate( shape=(sim.model.nvar,N,buf_len) ) * gf]
        for i, svar_buf in enumerate(svar_bufs):
            svar_buf[:,:horizon] = np.roll(sim.history.buffer[:,i,:,0], -1, axis=0).T

        if sim.stimulus is None:
            stimulus = None
        else:
            sim.stimulus.configure_space()
            sim.stimulus.configure_time(np.arange(nstep)*sim.integrator.dt)
            stimulus = sim.stimulus()

        svar_bufs = self._run_integrate( integrate, sim, N, nstep, svar_bufs, stimulus)

        return [svar_buf[:,horizon:] for svar_buf in svar_bufs]


    def _run_integrate(self, integrate, sim, N, nstep, svar_bufs, stimulus):
        return integrate(
            N = N,
            dt = sim.integrator.dt,
            nstep = nstep,
            i0 = sim.connectivity.horizon,
            **dict(zip(sim.model.state_variables,svar_bufs)),
            weights = sim.connectivity.weights, 
            idelays = sim.connectivity.idelays,
            parmat = sim.model.spatial_parameter_matrix.T,
            stimulus = stimulus
        )

    def _time_average(self, ts, istep):
        N, T = ts.shape
        return np.mean(ts.reshape(N,T//istep,istep),-1) # length of ts better be multiple of istep 

    def _run_sim_tavg_chunked(self, sim, nstep, chunksize, compatibility_mode=False, print_source=False):
        template = '<%include file="nb-montbrio.py.mako"/>'
        content = dict(sim=sim, compatibility_mode=compatibility_mode) 
        integrate = self.build_py_func(template, content, name='integrate', print_source=print_source)
        # chunksize in number of steps 
        horizon = sim.connectivity.horizon
        N = sim.connectivity.number_of_regions
        gf = sim.integrator.noise.gfun(None)

        tavg_steps=sim.monitors[0].istep
        assert tavg_steps < chunksize
        assert chunksize % tavg_steps == 0
        tavg_chunksize = chunksize // tavg_steps


        assert nstep % tavg_steps == 0
        svar_outs = [svar_out for svar_out in np.zeros((sim.model.nvar,N,nstep//tavg_steps))]

        svar_bufs = [buf for buf in sim.integrator.noise.generate( shape=(sim.model.nvar,N,chunksize+horizon) ) * gf]
        for i, svar_buf in enumerate(svar_bufs):
            svar_buf[:,:horizon] = np.roll(sim.history.buffer[:,i,:,0], -1, axis=0).T


        for chunk, _ in enumerate(range(horizon, nstep+horizon, chunksize)):
            if sim.stimulus is None:
                stimulus = None
            else:
                sim.stimulus.configure_space()
                sim.stimulus.configure_time(
                        np.arange(chunk*chunksize, (chunk+1)*chunksize)*sim.integrator.dt
                )
                stimulus = sim.stimulus()
            svar_bufs = self._run_integrate( integrate, sim, N, chunksize, svar_bufs, stimulus)

            tavg_chunk = chunk * tavg_chunksize
            svar_chunks = [self._time_average(svar[:, horizon:], tavg_steps) for svar in svar_bufs]
            if tavg_chunk+tavg_chunksize > svar_outs[0].shape[1]:
                cutoff = tavg_chunk+tavg_chunksize - svar_outs[0].shape[1]
                for svar_chunk in svar_chunks:
                    svar_chunk = svar_chunk[:,:-cutoff]
                    
            for svar_out, svar_chunk in zip (svar_outs, svar_chunks):
                svar_out[:,tavg_chunk:tavg_chunk+tavg_chunksize] = svar_chunk

            for svar_buf in svar_bufs:
                svar_buf[:,:horizon] = svar_buf[:,-horizon:]


            for svar_buf, svar_noise in zip(svar_bufs, sim.integrator.noise.generate( shape=(sim.model.nvar,N,chunksize) ) * gf):
                svar_buf[:,horizon:] = svar_noise


        return svar_outs



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
The TimeSeries datatypes.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

from io import BytesIO
from tvb.datatypes import sensors, surfaces, volumes, region_mapping, connectivity
from tvb.basic.neotraits.api import HasTraits, Attr, NArray, List, Float, narray_summary_info
from tvb.basic.readers import H5Reader
import numpy
from copy import deepcopy


class TimeSeries(HasTraits):
    """
    Base time-series dataType.
    """
    title = Attr(str)

    data = NArray(
        label="Time-series data",
        doc="""An array of time-series data, with a shape of [tpts, :], where ':' represents 1 or more dimensions""")

    labels_ordering = List(
        default=("Time", "State Variable", "Space", "Mode"),
        label="Dimension Names",
        doc="""List of strings representing names of each data dimension""")

    labels_dimensions = Attr(
        field_type=dict,
        default={},
        label="Specific labels for each dimension for the data stored in this timeseries.",
        doc=""" A dictionary containing mappings of the form {'dimension_name' : [labels for this dimension] }""")

    time = NArray(
        label="Time-series time",
        required=False,
        doc="""An array of time values for the time-series, with a shape of [tpts,].
            This is 'time' as returned by the simulator's monitors.""")

    start_time = Float(label="Start Time:")

    sample_period = Float(label="Sample period", default=1.0)

    # Specify the measure unit for sample period (e.g sec, msec, usec, ...)
    sample_period_unit = Attr(
        field_type=str,
        label="Sample Period Measure Unit",
        default="ms")

    @property
    def nr_dimensions(self):
        return self.data.ndim

    @property
    def sample_rate(self):
        """:returns samples per second [Hz] """
        if self.sample_period_unit in ("s", "sec"):
            return 1.0 / self.sample_period
        elif self.sample_period_unit in ("ms", "msec"):
            return 1000.0 / self.sample_period
        elif self.sample_period_unit in ("us", "usec"):
            return 1000000.0 / self.sample_period
        else:
            raise ValueError(f"{self.sample_period_unit} is not a recognized time unit")

    @property
    def sample_period_ms(self):
        """:returns sample_period is ms """
        if self.sample_period_unit in ("s", "sec"):
            return 1000 * self.sample_period
        elif self.sample_period_unit in ("ms", "msec"):
            return self.sample_period
        elif self.sample_period_unit in ("us", "usec"):
            return self.sample_period / 1000.0
        else:
            raise ValueError(f"{self.sample_period_unit} is not a recognized time unit")

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = {
            "Time-series type": self.__class__.__name__,
            "Time-series name": self.title,
            "Dimensions": self.labels_ordering,
            "Time units": self.sample_period_unit,
            "Sample period": self.sample_period,
            "Start time": self.start_time,
            "Length": self.sample_period * self.data.shape[0]
        }
        summary.update(narray_summary_info(self.data))
        return summary

    def duplicate(self, **kwargs):
        duplicate = super(TimeSeries, self).duplicate()
        for attr, value in kwargs.items():
            setattr(duplicate, attr, value)
        duplicate.configure()
        return duplicate

    def _get_index_of_state_variable(self, sv_label):
        try:
            sv_index = numpy.where(self.variables_labels == sv_label)[0][0]
        except KeyError:
            self.logger.error("There are no state variables defined for this instance. Its shape is: %s",
                              self.data.shape)
            raise
        except IndexError:
            self.logger.error("Cannot access index of state variable label: %s. Existing state variables: %s" % (
                sv_label, self.variables_labels))
            raise
        return sv_index

    def get_state_variable(self, sv_label):
        sv_data = self.data[:, self._get_index_of_state_variable(sv_label), :, :]
        subspace_labels_dimensions = deepcopy(self.labels_dimensions)
        subspace_labels_dimensions[self.labels_ordering[1]] = [sv_label]
        if sv_data.ndim == 3:
            sv_data = numpy.expand_dims(sv_data, 1)
        return self.duplicate(data=sv_data, labels_dimensions=subspace_labels_dimensions)

    def _get_indices_for_labels(self, list_of_labels):
        list_of_indices_for_labels = []
        for label in list_of_labels:
            try:
                space_index = numpy.where(self.space_labels == label)[0][0]
            except ValueError:
                self.logger.error("Cannot access index of space label: %s. Existing space labels: %s" %
                                  (label, self.space_labels))
                raise
            list_of_indices_for_labels.append(space_index)
        return list_of_indices_for_labels

    def get_subspace_by_index(self, list_of_index, **kwargs):
        self._check_space_indices(list_of_index)
        subspace_data = self.data[:, :, list_of_index, :]
        subspace_labels_dimensions = deepcopy(self.labels_dimensions)
        subspace_labels_dimensions[self.labels_ordering[2]] = self.space_labels[list_of_index].tolist()
        if subspace_data.ndim == 3:
            subspace_data = numpy.expand_dims(subspace_data, 2)
        return self.duplicate(data=subspace_data, labels_dimensions=subspace_labels_dimensions, **kwargs)

    def get_subspace_by_labels(self, list_of_labels):
        list_of_indices_for_labels = self._get_indices_for_labels(list_of_labels)
        return self.get_subspace_by_index(list_of_indices_for_labels)

    def __getattr__(self, attr_name):
        if self.labels_ordering[1] in self.labels_dimensions.keys():
            if attr_name in self.variables_labels:
                return self.get_state_variable(attr_name)
        if self.labels_ordering[2] in self.labels_dimensions.keys():
            if attr_name in self.space_labels:
                return self.get_subspace_by_labels([attr_name])
        raise AttributeError("%r object has no attribute %r" % (self.__class__.__name__, attr_name))

    def _get_index_for_slice_label(self, slice_label, slice_idx):
        if slice_idx == 1:
            return self._get_indices_for_labels([slice_label])[0]
        if slice_idx == 2:
            return self._get_index_of_state_variable(slice_label)

    @property
    def shape(self):
        return self.data.shape

    @property
    def time_unit(self):
        return self.sample_period_unit

    @property
    def space_labels(self):
         return numpy.array(self.labels_dimensions.get(self.labels_ordering[2], []))

    @property
    def variables_labels(self):
        return numpy.array(self.labels_dimensions.get(self.labels_ordering[1], []))

    def _check_space_indices(self, list_of_index):
        for index in list_of_index:
            if index < 0 or index > self.data.shape[1]:
                self.logger.error("Some of the given indices are out of space range: [0, %s]",
                                  self.data.shape[1])
                raise IndexError

    @classmethod
    def from_bytes_stream(cls, bytes_stream, content_type=".npz"):
        result = TimeSeries()

        if content_type == '.npz':
            ts_data = numpy.load(BytesIO(bytes_stream))
            result.data = ts_data['data']
            result.time = ts_data['time']
            return result

        reader = H5Reader(BytesIO(bytes_stream))
        result.data = reader.read_field("data")
        result.time = reader.read_optional_field("time")
        return result


class SensorsTSBase(TimeSeries):

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(SensorsTSBase, self).summary_info()
        summary.update({"Source Sensors": self.sensors.title})
        return summary


class TimeSeriesEEG(SensorsTSBase):
    """ A time series associated with a set of EEG sensors. """

    sensors = Attr(field_type=sensors.SensorsEEG)
    labels_ordering = List(of=str, default=("Time", "SV", "EEG Sensor", "Mode"))


class TimeSeriesMEG(SensorsTSBase):
    """ A time series associated with a set of MEG sensors. """

    sensors = Attr(field_type=sensors.SensorsMEG)
    labels_ordering = List(of=str, default=("Time", "SV", "MEG Sensor", "Mode"))


class TimeSeriesSEEG(SensorsTSBase):
    """ A time series associated with a set of Internal sensors. """

    sensors = Attr(field_type=sensors.SensorsInternal)
    labels_ordering = List(of=str, default=("Time", "SV", "sEEG Sensor", "Mode"))


class TimeSeriesRegion(TimeSeries):
    """ A time-series associated with the regions of a connectivity. """

    connectivity = Attr(field_type=connectivity.Connectivity)
    region_mapping_volume = Attr(field_type=region_mapping.RegionVolumeMapping, required=False)
    region_mapping = Attr(field_type=region_mapping.RegionMapping, required=False)
    labels_ordering = List(of=str, default=("Time", "State Variable", "Region", "Mode"))

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(TimeSeriesRegion, self).summary_info()
        summary.update({
            "Source Connectivity": self.connectivity.title,
            "Region Mapping": self.region_mapping.title if self.region_mapping else "None",
            "Region Mapping Volume": (self.region_mapping_volume.title
                                      if self.region_mapping_volume else "None")
        })
        return summary


class TimeSeriesSurface(TimeSeries):
    """ A time-series associated with a Surface. """

    surface = Attr(field_type=surfaces.CorticalSurface)
    labels_ordering = List(of=str, default=("Time", "State Variable", "Vertex", "Mode"))

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(TimeSeriesSurface, self).summary_info()
        summary.update({"Source Surface": self.surface.title})
        return summary


class TimeSeriesVolume(TimeSeries):
    """ A time-series associated with a Volume. """

    volume = Attr(field_type=volumes.Volume)
    labels_ordering = List(of=str, default=("Time", "X", "Y", "Z"))

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(TimeSeriesVolume, self).summary_info()
        summary.update({"Source Volume": self.volume.title})
        return summary



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
A module of classes and functions of common use.

.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>
.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>
.. moduleauthor:: Paula Sanz Leon <Paula@tvb.invalid>

"""

import numpy
import os
import re
import six
import logging
from tvb.basic.logger.builder import GLOBAL_LOGGER_BUILDER, get_logger
from .backend.ref import ReferenceBackend
from deprecated import deprecated



@deprecated(reason="Use builder.set_logger_level instead")
def log_debug(debug=False, timestamp=False, prefix=''):
    level_name = 'DEBUG' if debug else 'INFO'
    level = getattr(logging, level_name)
    GLOBAL_LOGGER_BUILDER.set_loggers_level(level)
    LOG = get_logger(__name__)
    for handler in LOG.root.handlers:
        handler.setLevel(level)
        # reset formatter more friendly for console work
        if isinstance(handler, logging.StreamHandler) and not timestamp:
            if prefix:
                prefix += ' '
            handler.setFormatter(logging.Formatter(prefix + '%(levelname)07s  %(message)s'))
    LOG.info('log level set to %s' % (level_name,))


def astr(ary):
    """Make short str repr of numerical value."""
    if isinstance(ary, numpy.ndarray):
        if ary.size == 1:
            val = ary[0]
        else:
            val = 'ndarray(%s, %s)' % (ary.shape, ary.dtype)
    elif isinstance(ary, bool):
        val = str(ary)
    elif isinstance(ary, float) or isinstance(ary, six.integer_types):
        val = ary
    else:
        val = str(ary)

    if isinstance(val, str):
        return val
    else:
        is_py_int = isinstance(val, six.integer_types)
        is_np_int = hasattr(val, 'dtype') and numpy.issubdtype(ary.dtype, numpy.integer)
        if is_py_int or is_np_int:
            return '%d' % (val,)
        else:
            return '%g' % (val,)


def map_astr(self, names):
    """Helper for generating a sequence of astr representation of attributes on self"""
    strs = []
    for name in names.split():
        strs.append(astr(getattr(self, name)))
    return tuple(strs)


def simple_gen_astr(self, names):
    """Helper for generating str for object with only numerical attributes."""
    strs = []
    for name, str_ in zip(names.split(), map_astr(self, names)):
        strs.append('%s=%s' % (name, str_))
    clsname = self.__class__.__name__
    return '%s(%s)' % (clsname, ', '.join(strs))


numpy_add_at = ReferenceBackend.add_at

# loose couple psutil so it's an optional dependency
try:
    import psutil
except ImportError:
    msg = """psutil module not available: no warnings will be issued when a
    simulation may require more memory than available"""
    LOG = get_logger(__name__)
    LOG.warning(msg)
    psutil = None


class Struct(dict):
    """
    the Struct class is a dictionary with matlab/C struct-like access
    to its fields:

    >>> parameters = Struct(x=23.4345, alpha=1.522e-4)
    >>> parameters.x + 1
    24.4345
    >>> parameters.x_init = 6
    >>> parameters.x_init + 1
    7
    >>> print(parameters.y)
    None

    note that this class returns None if the field does not exist!

    """

    def __getattr__(self, attr):
        return self.get(attr, None)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


linear_interp1d = ReferenceBackend.linear_interp1d
heaviside = ReferenceBackend.heaviside
iround = ReferenceBackend.iround

def zip_directory(path, zip_file):
    """
    Zip a given directory...
    Didn't know where to put this. 
    I need to pack zips from the scripting interface. 
    To avoid duplicating code I leave this small function here. 

    :param path -- where to store the zip
    :param zip_file 
    """
    for dirname, subdirs, files in os.walk(path):
        zip_file.write(dirname)
        for filename in files:
            zip_file.write(os.path.join(dirname, filename))
        zip_file.close()


def total_ms(duration="", hours=0, minutes=0, seconds=0):
    re_times = re.compile(r'\s*(\d+\.?\d*)\s*(sec|ms|hr|min|s|m|h)\s*$')
    re_decimals = re.compile(r'^[0-9]+([,.][0-9]+)?$')
    total_milliseconds = 0
    if duration == "":
        if not hours == 0:
            total_milliseconds = hours * 3600000
        elif not minutes == 0:
            total_milliseconds = minutes * 60000
        elif not seconds == 0:
            total_milliseconds = seconds * 1000
    elif re_decimals.search(duration):
        total_milliseconds = float(duration)
    else:
        match = re_times.match(duration)
        if not match:
            raise ValueError("unable to parse duration %r" % duration)
        s_time, s_unit = match.group(1, 2)
        s_time = float(s_time)
        if s_unit in ('s', 'sec'):
            total_milliseconds = s_time * 1000
        elif s_unit in ('m', 'min'):
            total_milliseconds = s_time * 60000
        elif s_unit in ('hr', 'h'):
            total_milliseconds = s_time * 3600000
        elif s_unit in ('ms'):
            total_milliseconds = s_time
    return total_milliseconds



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#
"""
Created on Mar 20, 2013

.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

import numpy
from tvb.tests.library.base_testcase import BaseTestCase
from tvb.datatypes import mode_decompositions, time_series


class TestModeDecompositions(BaseTestCase):
    """
    Tests the defaults for `tvb.datatypes.mode_decompositions` module.
    """

    def test_principalcomponents(self):
        data = numpy.random.random((10, 10, 10, 10))
        ts = time_series.TimeSeries(data=data)
        dt = mode_decompositions.PrincipalComponents(source=ts,
                                                     fractions=numpy.random.random((10, 10, 10)),
                                                     weights=data)
        dt.compute_norm_source()
        dt.compute_component_time_series()
        dt.compute_normalised_component_time_series()
        dt.configure()
        summary = dt.summary_info()
        assert summary['Mode decomposition type'] == 'PrincipalComponents'
        assert dt.source is not None
        assert dt.weights.shape == (10, 10, 10, 10)
        assert dt.fractions.shape == (10, 10, 10)
        assert dt.norm_source.shape == (10, 10, 10, 10)
        assert dt.component_time_series.shape == (10, 10, 10, 10)
        assert dt.normalised_component_time_series.shape == (10, 10, 10, 10)

    def test_independentcomponents(self):
        data = numpy.random.random((10, 10, 10, 10))
        ts = time_series.TimeSeries(data=data)
        n_comp = 5
        dt = mode_decompositions.IndependentComponents(source=ts,
                                                       component_time_series=numpy.random.random((10, n_comp, 10, 10)),
                                                       prewhitening_matrix=numpy.random.random((n_comp, 10, 10, 10)),
                                                       unmixing_matrix=numpy.random.random((n_comp, n_comp, 10, 10)),
                                                       n_components=n_comp)
        dt.compute_norm_source()
        dt.compute_component_time_series()
        dt.compute_normalised_component_time_series()
        summary = dt.summary_info()
        assert summary['Mode decomposition type'] == 'IndependentComponents'
        assert dt.source is not None
        assert dt.mixing_matrix is None
        assert dt.unmixing_matrix.shape == (n_comp, n_comp, 10, 10)
        assert dt.prewhitening_matrix.shape == (n_comp, 10, 10, 10)
        assert dt.norm_source.shape == (10, 10, 10, 10)
        assert dt.component_time_series.shape == (10, 10, n_comp, 10)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
TVB DataTypes, as a dictionary between multiple algorithms.
"""

__all__ = ["connectivity", "equations", "fcd", "graph", "local_connectivity",
           "mode_decompositions", "patterns", "projections",
           "region_mapping", "sensors", "structural", "spectral", "surfaces",
           "temporal_correlations", "time_series", "tracts", "volumes"]



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""

The Temporal Correlation datatypes.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

import tvb.datatypes.time_series as time_series
from tvb.basic.neotraits.api import HasTraits, Attr, NArray, List, narray_summary_info


class CrossCorrelation(HasTraits):
    """
    Result of a CrossCorrelation Analysis.
    """
    array_data = NArray()

    source = Attr(
        field_type=time_series.TimeSeries,
        label="Source time-series",
        doc="""Links to the time-series on which the cross_correlation is applied."""
    )

    time = NArray(label="Temporal Offsets", required=False)

    labels_ordering = List(
        of=str,
        label="Dimension Names",
        default=("Offsets", "Node", "Node", "State Variable", "Mode"),
        doc="""List of strings representing names of each data dimension"""
    )

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = {
            "Temporal correlation type": self.__class__.__name__,
            "Source": self.source.title,
            "Dimensions": self.labels_ordering
        }
        summary.update(narray_summary_info(self.array_data))
        return summary



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
An interactive power spectra plot generated from a TVB TimeSeries datatype.

Usage
::

    #Load the demo data
    import numpy
    data = numpy.load("demos/demo_data_region_16s_2048Hz.npy")
    period = 0.00048828125 #NOTE: Providing period in seconds

    #Create a tvb TimeSeries object
    import tvb.datatypes.time_series
    tsr = tvb.datatypes.time_series.TimeSeriesRegion()
    tsr.data = data
    tsr.sample_period = period
    tsr.sample_period_unit = 's'

    #Create and launch the interactive visualiser
    import tvb.simulator.power_spectra_interactive as ps_int
    psi = ps_int.PowerSpectraInteractive()
    psi.time_series = tsr
    psi.show()


.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

import numpy
import matplotlib.pyplot as plt
import ipywidgets as widgets
from deprecated import deprecated
from IPython.core.display import display
from tvb.simulator.common import get_logger
import tvb.datatypes.time_series as time_series_datatypes
from tvb.basic.neotraits.api import HasTraits, Attr, Int
from tvb.simulator.plot.utils import generate_region_demo_data

LOG = get_logger(__name__)

# Define a colour theme... see: matplotlib.colors.cnames.keys()
BACKGROUNDCOLOUR = "white"
EDGECOLOUR = "darkslateblue"
AXCOLOUR = "steelblue"
BUTTONCOLOUR = "steelblue"
HOVERCOLOUR = "blue"


@deprecated(reason="Use from tvb.contrib.scripts.plot PowerSpectraCoherenceInteractive")
class PowerSpectraInteractive(HasTraits):
    """
    The graphical interface for visualising the power-spectra (FFT) of a
    timeseries provide controls for setting:

        - which state-variable and mode to display [sets]
        - log or linear scaling for the power or frequency axis [binary]
        - segmentation length [set]
        - windowing function [set]
        - power normalisation [binary] (emphasise relative frequency contribution)
        - show std or sem [binary]


    """

    time_series = Attr(
        field_type=time_series_datatypes.TimeSeries,
        label="Timeseries",
        default=None,
        doc=""" The timeseries to which the FFT is to be applied.""")

    first_n = Int(
        label="Display the first 'n'",
        default=-1,
        doc="""Primarily intended for displaying the first N components of a 
            surface PCA timeseries. Defaults to -1, meaning it'll display all
            of 'space' (ie, regions or vertices or channels). In other words,
            for Region or M/EEG timeseries you can ignore this, but, for a 
            surface timeseries it really must be set.""")

    def __init__(self, **kwargs):
        """
        Initialise based on provided keywords or their traited defaults. Also,
        initialise the place-holder attributes that aren't filled until the
        show() method is called.

        """
        super(PowerSpectraInteractive, self).__init__(**kwargs)
        LOG.debug(str(kwargs))
        # figure
        self.fig = None

        # time-series
        self.fft_ax = None

        # Current state
        self.xscale = "linear"
        self.yscale = "log"
        self.mode = 0
        self.variable = 0
        self.show_sem = False
        self.show_std = False
        self.normalise_power = "no"
        self.window_length = 0.25
        self.window_function = "None"

        # Selectors
        self.xscale_selector = None
        self.yscale_selector = None
        self.mode_selector = None
        self.variable_selector = None
        self.show_sem_selector = None
        self.show_std_selector = None
        self.normalise_power_selector = None
        self.window_length_selector = None
        self.window_function_selector = None

        #
        possible_freq_steps = [2 ** x for x in range(-2, 7)]  # Hz
        self.possible_window_lengths = 1.0 / numpy.array(possible_freq_steps)  # s
        self.freq_step = 1.0 / self.window_length
        self.frequency = None
        self.spectra = None
        self.spectra_norm = None

    def configure(self):
        """ Separate configure cause traits be busted... """
        LOG.debug("time_series shape: %s" % str(self.time_series.data.shape))
        self.data = self.time_series.data[:, :, :self.first_n, :]
        self.period = 1 / self.time_series.sample_rate
        self.period_unit = "s"
        self.max_freq = 0.5 / self.period
        self.units = "Hz"
        self.tpts = self.data.shape[0]
        self.nsrs = self.data.shape[2]
        self.time_series_length = self.tpts * self.period
        self.time = numpy.arange(self.tpts) * self.period
        self.labels = ["channel_%0.3d" % k for k in range(self.nsrs)]

    def show(self):
        """ Generate the interactive power-spectra figure. """
        # Make sure everything is configured
        self.configure()

        # Make the figure:
        fig = self.create_figure()

        self.calc_fft()
        self.plot_spectra()

        display(fig)

    def add_selectors_widgets_to_lc_items(self):
        self.add_mode_selector()
        self.add_normalise_power_selector()

        self.add_xscale_selector()
        self.add_yscale_selector()

        self.add_variable_selector()

        self.add_window_function_selector()
        self.add_window_length_selector()

        self.mode_sv_box = widgets.VBox([self.ms_box, self.vs_box], layout=self.outer_box_layout)
        self.xs_ys_box = widgets.VBox([self.xss_box, self.yss_box], layout=self.outer_box_layout)
        self.np_box = widgets.VBox([self.nps_box], layout=self.outer_box_layout)
        self.wf_box = widgets.VBox([self.wfs_box], layout=self.outer_box_layout)
        self.wl_box = widgets.VBox([self.wls_box], layout=self.outer_box_layout)

        self.lc_items.extend([self.mode_sv_box, self.xs_ys_box, self.np_box, self.wf_box, self.wl_box])

    ##------------------------------------------------------------------------##
    ##------------------ Functions for building the figure -------------------##
    ##------------------------------------------------------------------------##
    def create_figure(self):
        """ Create the figure and time-series axes. """

        self.outer_box_layout = widgets.Layout(border='solid 1px white',
                                               margin='3px 3px 3px 3px',
                                               padding='2px 2px 2px 2px')

        self.box_layout = widgets.Layout(border='solid 1px black',
                                         margin='3px 3px 3px 3px',
                                         padding='2px 2px 2px 2px')

        self.other_layout = widgets.Layout(width='90%')

        self.lc_items = []

        self.add_selectors_widgets_to_lc_items()

        self.lc_box = widgets.HBox(self.lc_items)
        self.lc_box.layout = self.box_layout

        # item 2
        self.fig = None
        self.op = widgets.Output()
        self.op.layout = self.box_layout

        items = [self.op, self.lc_box]
        grid = widgets.GridBox(items, layout=widgets.Layout(grid_template_rows="450px 250px"))
        return grid

    def add_xscale_selector(self):
        """
        Add a radio button to the figure for selecting which scaling the x-axes
        should use.
        """
        xscale_tuple = ("log", "linear")
        self.xscale_selector = widgets.RadioButtons(options=xscale_tuple, value=xscale_tuple[1],
                                                    layout=self.other_layout)
        self.xscale_selector.observe(self.update_xscale, 'value')

        self.xss_box = widgets.VBox([widgets.Label('X Axis Scale'), self.xscale_selector], layout=self.box_layout)

    def add_yscale_selector(self):
        """
        Add a radio button to the figure for selecting which scaling the y-axes
        should use.
        """
        yscale_tuple = ("log", "linear")
        self.yscale_selector = widgets.RadioButtons(options=yscale_tuple, value=yscale_tuple[0],
                                                    layout=self.other_layout)
        self.yscale_selector.observe(self.update_yscale, 'value')

        self.yss_box = widgets.VBox([widgets.Label('Y Axis Scale'), self.yscale_selector], layout=self.box_layout)

    def add_mode_selector(self):
        """
        Add a radio button to the figure for selecting which mode of the model
        should be displayed.
        """
        mode_tuple = tuple(range(self.data.shape[3]))
        self.mode_selector = widgets.RadioButtons(options=mode_tuple, value=mode_tuple[0], layout=self.other_layout)
        self.mode_selector.observe(self.update_mode, 'value')

        self.ms_box = widgets.VBox([widgets.Label('Mode'), self.mode_selector], layout=self.box_layout)

    def add_variable_selector(self):
        """
        Generate radio selector buttons to set which state variable is 
        displayed.
        """
        noc = self.data.shape[1]
        self.variable_selector = widgets.RadioButtons(options=tuple(range(noc)), value=0, layout=self.other_layout)
        self.variable_selector.observe(self.update_variable, 'value')

        self.vs_box = widgets.VBox([widgets.Label('State Variable'), self.variable_selector], layout=self.box_layout)

    def add_window_length_selector(self):
        """
        Generate radio selector buttons to set the window length is seconds.
        """
        wl_tup = tuple(self.possible_window_lengths)
        self.window_length_selector = widgets.RadioButtons(options=wl_tup, value=wl_tup[4], layout=self.other_layout)
        self.window_length_selector.observe(self.update_window_length, 'value')

        self.wls_box = widgets.VBox([widgets.Label('Segment Length'), self.window_length_selector],
                                    layout=self.box_layout)

    def add_window_function_selector(self):
        """
        Generate radio selector buttons to set the windowing function.
        """
        wf_tup = ("None", "hamming", "bartlett", "blackman", "hanning")
        self.window_function_selector = widgets.RadioButtons(options=wf_tup, value=wf_tup[0], layout=self.other_layout)
        self.window_function_selector.observe(self.update_window_function, 'value')

        self.wfs_box = widgets.VBox([widgets.Label('Windowing Function'), self.window_function_selector],
                                    layout=self.box_layout)

    def add_normalise_power_selector(self):
        """
        Add a radio button to chose whether or not the power of all spectra 
        should be normalised to 1.
        """
        np_tuple = ("yes", "no")
        self.normalise_power_selector = widgets.RadioButtons(options=np_tuple, value=np_tuple[1],
                                                             layout=self.other_layout)
        self.normalise_power_selector.observe(self.update_normalise_power, 'value')

        self.nps_box = widgets.VBox([widgets.Label('Normalise'), self.normalise_power_selector], layout=self.box_layout)

    ##------------------------------------------------------------------------##
    ##------------------ Functions for updating the state --------------------##
    ##------------------------------------------------------------------------##
    def calc_fft(self):
        """
        Calculate FFT using current state of the window_length, window_function,
        """
        # Segment time-series, overlapping if necessary
        nseg = int(numpy.ceil(self.time_series_length / self.window_length))
        if nseg != 1:
            seg_tpts = numpy.ceil(self.window_length / self.period)  # use ceil to avoid dimensions mismatch
            overlap = ((seg_tpts * nseg) - self.tpts) / (nseg - 1)
            starts = [max(seg * (seg_tpts - overlap), 0) for seg in range(nseg)]
            segments = [self.data[int(start):int(start + seg_tpts)] for start in starts]
            segments = [segment[:, :, :, numpy.newaxis] for segment in segments]
            time_series = numpy.concatenate(segments, axis=4)
        else:
            time_series = self.data[:, :, :, :, numpy.newaxis]
            seg_tpts = time_series.shape[0]

        # Base-line correct segmented time-series
        time_series = time_series - time_series.mean(axis=0)[numpy.newaxis, :]

        # Apply windowing function
        if self.window_function != "None":
            window_function = eval("".join(("numpy.", self.window_function)))
            window_mask = numpy.reshape(window_function(seg_tpts),
                                        (int(seg_tpts), 1, 1, 1, 1))
            time_series = time_series * window_mask

        # Calculate the FFT
        result = numpy.fft.fft(time_series, axis=0)
        nfreq = numpy.ceil(len(result) / 2)  # use ceil to avoid dimensions mismatch

        self.frequency = numpy.arange(0, self.max_freq, self.freq_step)
        LOG.debug("frequency shape: %s" % str(self.frequency.shape))

        self.spectra = numpy.mean(numpy.abs(result[1:int(nfreq) + 1]) ** 2, axis=-1)
        LOG.debug("spectra shape: %s" % str(self.spectra.shape))

        self.spectra_norm = (self.spectra / numpy.sum(self.spectra, axis=0))
        LOG.debug("spectra_norm shape: %s" % str(self.spectra_norm.shape))

    ##------------------------------------------------------------------------##
    ##------------------ Functions for updating the figure -------------------##
    ##------------------------------------------------------------------------##
    def update_xscale(self, value):
        """ 
        Update the FFT axes' xscale to either log or linear based on radio
        button selection.
        """
        self.xscale = value['new']
        self.fft_ax.set_xscale(self.xscale)
        plt.draw()

    def update_yscale(self, value):
        """ 
        Update the FFT axes' yscale to either log or linear based on radio
        button selection.
        """
        self.yscale = value['new']
        self.fft_ax.set_yscale(self.yscale)
        plt.draw()

    def update_mode(self, value):
        """ Update the visualised mode based on radio button selection. """
        self.mode = value['new']
        self.plot_spectra()

    def update_variable(self, value):
        """ 
        Update state variable being plotted based on radio button selection.
        """
        self.variable = value['new']
        self.plot_spectra()

    def update_normalise_power(self, value):
        """ Update whether to normalise based on radio button selection. """
        self.normalise_power = value['new']
        self.plot_spectra()

    def update_window_length(self, value):
        """
        Update timeseries window length based on the selected value.
        """
        self.window_length = numpy.float64(value['new'])
        self.freq_step = 1.0 / self.window_length
        self.update_spectra()

    def update_window_function(self, value):
        """
        Update windowing function based on the radio button selection.
        """
        self.window_function = value['new']
        self.update_spectra()

    def update_spectra(self):
        """ Clear the axes and redraw the power-spectra. """
        self.calc_fft()
        self.plot_spectra()

    def add_axes(self):
        self.fft_ax = self.fig.add_axes([0.15, 0.15, 0.75, 0.75])

    def plot_figure(self):
        if not self.fig:
            time_series_type = self.time_series.__class__.__name__
            try:
                figure_window_title = "Interactive power spectra: " + time_series_type
                plt.close(figure_window_title)
                self.fig = plt.figure(num=figure_window_title,
                                      figsize=(9, 4),
                                      facecolor=BACKGROUNDCOLOUR,
                                      edgecolor=EDGECOLOUR)
            except ValueError:
                LOG.info("My life would be easier if you'd update your Pyplot...")
                figure_number = 42
                plt.close(figure_number)
                self.fig = plt.figure(num=figure_number,
                                      figsize=(9, 4),
                                      facecolor=BACKGROUNDCOLOUR,
                                      edgecolor=EDGECOLOUR)
            self.add_axes()

    def fft_ax_plot(self):
        # import pdb; pdb.set_trace()
        # Plot the power spectra
        if self.normalise_power == "yes":
            self.fft_ax.plot(self.frequency,
                             self.spectra_norm[:, self.variable, :, self.mode])
        else:
            self.fft_ax.plot(self.frequency,
                             self.spectra[:, self.variable, :, self.mode])

    def plot_fft(self):
        self.fft_ax.clear()
        # Set title and axis labels
        time_series_type = self.time_series.__class__.__name__
        self.fft_ax.set(title=time_series_type)
        self.fft_ax.set(xlabel="Frequency (%s)" % self.units)
        self.fft_ax.set(ylabel="Power")

        # Set x and y scale based on curent radio button selection.
        self.fft_ax.set_xscale(self.xscale)
        self.fft_ax.set_yscale(self.yscale)

        if hasattr(self.fft_ax, 'autoscale'):
            self.fft_ax.autoscale(enable=True, axis='both', tight=True)

        self.fft_ax_plot()

    def plot_figures(self):
        self.plot_fft()

    def plot_spectra(self):
        """ Plot the power spectra. """
        self.op.clear_output(wait=True)
        with plt.ioff():
            if not self.fig:
                time_series_type = self.time_series.__class__.__name__
                try:
                    figure_window_title = "Interactive power spectra: " + time_series_type
                    plt.close(figure_window_title)
                    self.fig = plt.figure(num=figure_window_title,
                                          figsize=(9, 4),
                                          facecolor=BACKGROUNDCOLOUR,
                                          edgecolor=EDGECOLOUR)
                except ValueError:
                    LOG.info("My life would be easier if you'd update your Pyplot...")
                    figure_number = 42
                    plt.close(figure_number)
                    self.fig = plt.figure(num=figure_number,
                                          figsize=(9, 4),
                                          facecolor=BACKGROUNDCOLOUR,
                                          edgecolor=EDGECOLOUR)
            self.add_axes()
            self.plot_figures()

        with self.op:
            display(self.fig.canvas)


def main_function(class_type=PowerSpectraInteractive):
    import os
    # Do some stuff that tests or makes use of this module...
    LOG.info("Testing %s module..." % __file__)
    file_path = os.path.join(os.getcwd(), "demo_data_region_16s_2048Hz.npy")
    try:
        data = numpy.load(file_path)
    except IOError:
        LOG.error("Can't load demo data. It will be created now running the generate_region_demo_data() function.")
        generate_region_demo_data(file_path=file_path)
        data = numpy.load(file_path)

    period = 0.00048828125  # NOTE: Providing period in seconds
    tsr = time_series_datatypes.TimeSeriesRegion()
    tsr.data = data
    tsr.sample_period = period
    tsr.sample_period_unit = 's'

    psi = PowerSpectraInteractive()
    psi.time_series = tsr
    psi.show()


if __name__ == "__main__":
    main_function()
